import os
import nltk
import argparse
import time
import json

from nltk.corpus import framenet

from text2story.core.narrative import Narrative

PARTICIPANT_TYPES = ["focal_participant","participant","participant_1",\
        "participant_2","protagonist","victim","agent","instrument",\
        "theme","leader","name","member","creator","buyer",\
        "individuals","author","conqueror","sleeper","goal",\
        "weapon","resource","sent_item","owner","possessor",\
        "experiencer","user","student","substance"]

OTHER_ELEMENTS_TYPE = ["location"]

#POSSIBLE_PARTIPANTS = ["conqueror","medium","figure","cognizer","sleeper"]

def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")

class ReadFramenet:

    def __init__(self):
        nltk.download("framenet_v17")

    def convert_to_brat(self, lang, outputdir):

        self.build_txt_files(outputdir)
        self.build_ann_files(lang, outputdir)

    def add_arg_to_narrative(self, arg, narrative):

        # events
        event_offset_set = set()


        # it gets the event of the current frame
        event_txt = ""
        for (start, end) in arg.Target:

            event_offset_set.add((start, end))
            tok_txt = arg.text[start:end]
            event_txt = event_txt + tok_txt + " "

        event_txt = event_txt.rstrip()
        offset_start = narrative.text.find(event_txt, 0)
        event_id = None
        if offset_start < 0:
            # probably it is a non continous event
            print("Warning: Event index not located.")
        else:
            offset_end = offset_start + len(event_txt)
            event_id = narrative._add_event((offset_start, offset_end))


        # coletar aqui as relacoes, existe o type_relation
        # o temos acima o evento que conecta os atores abaixo. certo?
        # participants (FE)
        with open("fn-lirics.json","r") as fd:
            map_fn_lirics = json.load(fd)

        current_offset = 0
        for (start, end, type_relation) in arg.FE[0]:

            # it is an event already
            if (start,end) in event_offset_set:
                continue

            tok_txt = arg.text[start:end]
            offset_start = narrative.text.find(tok_txt, 0)
            offset_end = offset_start + len(tok_txt)
            current_offset = offset_end

            type_relation_low = type_relation.lower()


            # it is a type of event
            if type_relation_low.endswith("event") or\
                type_relation_low.startswith("event")    :
                event_id = narrative._add_event((offset_start, offset_end))
            else:

                if type_relation_low == "time":
                    time_id = narrative._add_time((offset_start, offset_end))

                if type_relation_low in OTHER_ELEMENTS_TYPE:
                    continue

                actor_id = None
                # it is an participant that is the participant type group
                if type_relation_low in PARTICIPANT_TYPES:
                    actor_id = narrative._add_actor((offset_start, offset_end))

                    # it is participant that can be mapped to the lirics type
                    # nao necessariamente, a condicao acima e a condicao abaixo devem ocorrer. Certo?
                    if type_relation_low in map_fn_lirics:
                        narrative._add_srlink(actor_id, event_id, type_relation_low)



    def build_ann_files(self, lang, outputdir):

        # ignore the annotations of exemplars since they are just 
        # examples, and we want to extract narrative from full text files
        ann_lst = framenet.annotations(exemplars=False)
        doc_lst = framenet.docs()

        # iterate through annotations, and map each document to a 
        # a narrative object
        map_narrative_doc = {}
        for ann in ann_lst:

            # get the document associated with  this annotation
            docid = ann.sent.docID
            doc = framenet.doc(docid)

            filename = os.path.splitext(doc.filename)[0] + ".txt"


            # it gets or creates the narrative object associated with this 
            # annotation 
            if filename in map_narrative_doc:
                narrative_inst = map_narrative_doc[filename]
            else:
                full_filename = os.path.join(outputdir, filename)

                txt = ""
                with open(full_filename, "r") as fd:
                    txt = fd.read()

                narrative_inst = Narrative(lang, txt,"0000-00-00")
                narrative_inst.text = txt

                map_narrative_doc[filename] = narrative_inst


            self.add_arg_to_narrative(ann, narrative_inst)

            filename_ann = os.path.splitext(doc.filename)[0] + ".ann"
            full_filename_ann = os.path.join(outputdir, filename_ann)

            narrative_ann_str = narrative_inst.ISO_annotation()

            with open(full_filename_ann, "w") as fd:
                fd.write(narrative_ann_str)
                


    def build_txt_files(self, outputdir):

        doc_lst = framenet.docs()

        for doc in doc_lst:

            doc_txt = ""
            for sent in doc.sentence:
                doc_txt = doc_txt + sent.text

            filename = os.path.splitext(doc.filename)[0] + ".txt"
            filename = os.path.join(outputdir, filename)

            with open(filename, "w") as fd:
                fd.write(doc_txt)


def main(lang, outputdir):

    reader = ReadFramenet()
    reader.convert_to_brat(lang, outputdir)

if __name__ == '__main__':
    my_parser = argparse.ArgumentParser(description='Convert Framenet corpus to brat format')

    my_parser.add_argument("language", action='store', type=str, help="The language to convert the framenet (pt or en).")
    my_parser.add_argument("datadir", action='store', type=dir_path, help="The directory where the brat files will be stored.")

    args = my_parser.parse_args()
    main(args.language, args.datadir)



# gather corpus stats
import os
import sys

from pathlib import Path
from text2story.readers.read_brat import  ReadBrat
from nltk import word_tokenize

import text2story as t2s
from text2story.core.exceptions import *
from text2story.core.entity_structures import *
from text2story.select.event import get_nested_events_reporting, get_embedded_events


def count_tokens(fd):
    return len(word_tokenize(fd.read()))

def count_elements(fd, element_type):

    count = 0
    for line in fd:
        if line[0] == "T":
            elements = line.split()
            if elements[1] == element_type:
                count += 1
        elif line[0] == "R":
            elements = line.split()
            if elements[1].startswith(element_type):
                count += 1

    return count

def get_attr_repr(attr_value):
    if type(attr_value) == list:
        return repr(attr_value)
    else:
        return attr_value

def build_stats_byelement(doc,  element_type="event", attr_filter={}, allowed_by_id=[]):

    attr_count = {}
    event_id_lst = set()
    allowed_id_lst = [id for (id, _) in allowed_by_id]

    for tok in doc:

        for ann_type, attr_map in tok.attr:


            if ann_type.lower() == element_type:
                if tok.id_ann[0] not in event_id_lst:
                    filter_flag = False
                    for attr in attr_map:
                        # check if the attribute satistifies a given filter
                        if attr in attr_filter and attr_map[attr] == attr_filter[attr]:
                            if tok.id_ann[0] not in allowed_id_lst:
                                filter_flag = True
                                break # ignore this element if it satisfies the requirement

                        attr_map_value = get_attr_repr(attr_map[attr])

                        if attr in attr_count:
                            attr_count[attr][attr_map_value] = attr_count[attr].get(attr_map_value, 0) + 1
                        else:
                            attr_count[attr] = {attr_map_value:1}

                    if not(filter_flag):
                       # if element_type == "quantificacao" and ann_type.lower() == "quantificacao":
                        #    print("-->", tok.id_ann[0], tok.text, tok.attr)
                        event_id_lst.add(tok.id_ann[0])


    return attr_count, len(event_id_lst)

def build_stats_byrelation(doc):
    rel_stats = {}
    rel_set = set()

    for tok in doc:
        for rel in tok.relations:
            if rel.rel_id not in rel_set:

                rel_type = rel.rel_type.split("_")
                maintype = rel_type[0]
                subtype = rel_type[1]

                if maintype in rel_stats:
                    if subtype in rel_stats[maintype]:
                        rel_stats[maintype][subtype] += 1
                    else:
                        rel_stats[maintype][subtype] = 1
                else:
                    rel_stats[maintype] = {subtype:1}

    return rel_stats

def build_stats_byelement_batch(data_dir, element_type="event", dataset_type="BRAT"):
    """

    Count a given type element

    @param data_dir: A string of data directory
    @param dataset_type: default type assumes the BRAT standoff format
    @return: dict of {attribute:frequency}
    """
    reader = ReadBrat()
    doc_lst = reader.process(data_dir)
    attr_count = {}


    for doc in doc_lst:
        attr_count.update(build_stats_byelement(doc, element_type, dataset_type))

    return attr_count

def get_entity(narrative, id_entity):

    if id_entity in narrative.events:
        return narrative.events[id_entity]
    elif id_entity in narrative.participants:
        return narrative.participants[id_entity]
    elif id_entity in narrative.times:
        return narrative.times[id_entity]
    else:
        raise InvalidIDAnn(id_entity)


def build_stats(data_dir):


    counts = {"tokens":[],"events":[],"participants":[],"time":[], "olink":[],
              "srlink":[],"tlink":[],"slink":[],"alink":[],"qslink":[],"movelink":[]}

    count_docs = 0

    for dirpath, dirnames, filenames in os.walk(data_dir):
        for f in filenames:
            if f.endswith(".ann"):

                full_name = os.path.join(dirpath, f)
                p = Path(f)
                txt_file = os.path.join(data_dir,p.stem + ".txt")

                with open(txt_file, "r") as fd:
                    counts["tokens"].append(count_tokens(fd))

                with open(full_name, "r") as fd:
                    counts["events"].append(count_elements(fd, "Event"))

                with open(full_name, "r") as fd:

                    counts["participants"].append(count_elements(fd, "Participant"))
                with open(full_name, "r") as fd:
                    counts["time"].append(count_elements(fd, "Time"))

                with open(full_name, "r") as fd:
                    counts["olink"].append(count_elements(fd, "OLINK"))

                with open(full_name, "r") as fd:
                    counts["srlink"].append(count_elements(fd, "SRLINK"))

                with open(full_name, "r") as fd:
                    counts["tlink"].append(count_elements(fd, "TLINK"))

                with open(full_name, "r") as fd:
                    counts["slink"].append(count_elements(fd, "SLINK"))

                with open(full_name, "r") as fd:
                    counts["alink"].append(count_elements(fd, "ALINK"))

                with open(full_name, "r") as fd:
                    counts["qslink"].append(count_elements(fd, "QSLINK"))

                with open(full_name, "r") as fd:
                    counts["movelink"].append(count_elements(fd, "MOVELINK"))

                count_docs += 1

    print("Total #docs:", count_docs)
    print("Total #tokens:", sum(counts["tokens"]))
    print("Average #tokens:", sum(counts["tokens"])/len(counts["tokens"]))
    print("Total #events:", sum(counts["events"]))
    print("Total #participants:", sum(counts["participants"]))
    print("Total #time:", sum(counts["time"]))
    print("Total #olink:", sum(counts["olink"]))
    print("Total #srlink:", sum(counts["srlink"]))
    print("Total #tlink:", sum(counts["tlink"]))
    print("Total #slink:", sum(counts["slink"]))
    print("Total #alink:", sum(counts["alink"]))
    print("Total #qslink:", sum(counts["qslink"]))
    print("Total #movelink:", sum(counts["movelink"]))

def count_tlinks_embedded_events_sent(doc):
    """
        This method count all tlinks in the subset of embedded events in
        reporting events
        @param doc: a doc as TokenCorpus list
        @return: a dictionary representing a histogram
        """

    embedded_events_lst, embedded_events_ids, count_supressed_reporting = get_embedded_events(doc)

    hist = {}  # histogram of tlinks betwen reporting events and nested events
    relation_id_lst = set()


    # events nested into sent_id
    for sent_id in embedded_events_lst:
        reporting_event, embedded_events = embedded_events_lst[sent_id]
        for ee in embedded_events:
            for rel in ee.relations:

                # we are collecting relations in which:
                # 1) The type is TLINK
                # 2) the relation isn't analyzed yet
                # 3) the another argument is also an embedded event
                if rel.rel_type.startswith("TLINK") and\
                        rel.rel_id not in relation_id_lst and\
                        rel.toks[0].id_ann[0] in embedded_events_ids:
                    #print("-->", rel.rel_id, rel.rel_type, rel.toks[0].id_ann[0], ee.id_ann[0])

                    if rel.rel_type in hist:
                        hist[rel.rel_type] += 1
                    else:
                        hist[rel.rel_type] = 1
                    relation_id_lst.add(rel.rel_id)
    return hist, count_supressed_reporting


def count_tlinks_embbeded_events(doc):
    """
    This method count all tlinks in the subset of embedded events in
    reporting events
    @param doc: a doc as TokenCorpus list
    @return: a dictionary representing a histogram
    """

    #for tok in doc:
    #    print("-->", tok.text, tok.dep)

    event_lst, nested_events, nested_events_ids = get_nested_events_reporting(doc)
    hist = {}  # histogram of tlinks betwen reporting events and nested events
    event_id_lst = set()

    # events nested into event_id
    for event_id in nested_events:
        for nest_event in nested_events[event_id]:

            for rel in nest_event.relations:

                # we are collecting relations in which:
                # 1) The type is TLINK
                # 2) Its id isn't already checked
                # 3) Its second argument it is not a reporting event (represented by the keys of nested_events)
                # 4) Its second argument is also a nested event
                if rel.rel_type.startswith("TLINK") and \
                        rel.rel_id not in event_id_lst and \
                        rel.toks[0].id_ann[0] not in nested_events.keys() and \
                        rel.toks[0].id_ann[0] in nested_events_ids:

                    if rel.rel_type in hist:
                        hist[rel.rel_type] += 1
                    else:
                        hist[rel.rel_type] = 1
                    event_id_lst.add(rel.rel_id)
    return(hist)

if __name__ == "__main__":
    data_dir = sys.argv[1]
    option = sys.argv[2]
    if option == "-s":
        build_stats(data_dir)
    #elif option == "-tlinks":
    #    count_tlinks_byevent(data_dir)
    else:
        print("Use:")
        print("python stats.py data_dir -s")
        print("python stats.oy data_dir -tlinks")



from markdown_it.rules_inline import entity


def get_all_entities(tok_lst, ann_type_lst=["time"]):
    entity_lst = {}

    for i in range(len(tok_lst)):
        tok = tok_lst[i]

        for attr_item in tok.attr:

            ann_type = attr_item[0]
            attr_map = attr_item[1]

            if ann_type.lower() in ann_type_lst:

                for id_ann in tok.id_ann:

                    if id_ann in entity_lst:
                        entity_lst[id_ann].append(tok)
                    else:
                        entity_lst[id_ann] = [tok]
                break

    return entity_lst


from transformers import Pipeline, AutoModel, AutoTokenizer
import spacy
import spacy_transformers
from torch import Tensor
import torch
from text2story.annotators.SRL.evaluator import Evaluator
import logging
from string import punctuation, whitespace

logger = logging.getLogger(__name__)


# Study wether subclassing is worth it due to the needed changes in postprocessing and sanitize parameters
class SrlPipeline(Pipeline):
    def __init__(self, model: str, tokenizer: str, framework, task, **kwargs):
        super().__init__(model, tokenizer=tokenizer)
        if "lang" in kwargs and kwargs["lang"] == "en":
            logger.info("Loading English verb predictor model...")
            if not (spacy.util.is_package('en_core_web_lg')):
                spacy.cli.download('en_core_web_lg')
            self.verb_predictor = spacy.load("en_core_web_lg")
        else:
            logger.info("Loading Portuguese verb predictor model...")
            if not (spacy.util.is_package('pt_core_news_lg')):
                spacy.cli.download('pt_core_news_lg')
            self.verb_predictor = spacy.load("pt_core_news_lg")
        logger.info("Got verb prediction model\n")

    def _sanitize_parameters(self, **kwargs):
        preprocess_kwargs = {}
        if "maybe_arg" in kwargs:
            preprocess_kwargs["maybe_arg"] = kwargs["maybe_arg"]
        return preprocess_kwargs, {}, {}

    def preprocess(self, sentence: str):
        # Extract sentence verbs
        doc = self.verb_predictor(sentence)
        #print([token.pos_ for token in doc])
        verbs = [token.text for token in doc if token.pos_ == "VERB"]
        # If the sentence only contains auxiliary verbs, consider those as the main verbs
        if verbs == []:
            verbs = [token.text for token in doc if token.pos_ == "AUX"]
        #print(verbs)
        # Remove duplicates
        verbs = list(dict.fromkeys(verbs))

            # Tokenize sentence
        tokens = self.tokenizer.encode_plus(
            sentence,
            truncation=True,
            return_token_type_ids=False,
            return_offsets_mapping=True,
        )
        tokens_lst = tokens.tokens()
        offsets = tokens["offset_mapping"]

        input_ids = torch.tensor([tokens["input_ids"]], dtype=torch.long)
        attention_mask = torch.tensor([tokens["attention_mask"]], dtype=torch.long)
        # token_type_ids = torch.tensor([token_type_id], dtype=torch.long)

        model_input = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "token_type_ids": [],
            "tokens": tokens_lst,
            "verb": "",
        }

        #model_inputs = [
        #     {**model_input} for _ in verbs
        #]  # Create a new dictionary for each verb
        model_inputs = [
            {
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "token_type_ids": [],
                "tokens": tokens_lst,
                "verb": ""
            } for _ in verbs
        ]
        for i, verb in enumerate(verbs):
            model_inputs[i]["verb"] = verb
            token_type_ids = model_inputs[i]["token_type_ids"]
            token_type_ids.append([])
            curr_word_offsets: tuple[int, int] = None

            for j in range(len(tokens_lst)):
                curr_offsets = offsets[j]
                curr_slice = sentence[curr_offsets[0] : curr_offsets[1]]
                if not curr_slice:
                    token_type_ids[-1].append(0)
                # Check if new token still belongs to same word
                elif (
                    curr_word_offsets
                    and curr_offsets[0] >= curr_word_offsets[0]
                    and curr_offsets[1] <= curr_word_offsets[1]
                ):
                    # Extend previous token type
                    token_type_ids[-1].append(token_type_ids[-1][-1])
                else:
                    curr_word_offsets = self._find_word(sentence, start=curr_offsets[0])
                    curr_word = sentence[curr_word_offsets[0] : curr_word_offsets[1]]

                    token_type_ids[-1].append(
                        int(curr_word != "" and curr_word == verb)
                    )

            model_inputs[i]["token_type_ids"] = torch.tensor(
                token_type_ids, dtype=torch.long
            )

        #print(sentence, model_inputs)
        return model_inputs

    def _forward(self, model_inputs):
        outputs = []
        for model_input in model_inputs:
            output = self.model(
                input_ids=model_input["input_ids"],
                attention_mask=model_input["attention_mask"],
                token_type_ids=model_input["token_type_ids"],
            )
            #output["verb"] = model_input["verb"]
            #output["tokens"] = model_input["tokens"]
            output_dict = {
                "class_probabilities": output["class_probabilities"],  # ou output.class_probabilities
                "attention_mask": model_input["attention_mask"],  # ✅ Adicione isto!
                "verb": model_input["verb"],
                "tokens": model_input["tokens"]
            }
            outputs.append(output_dict)
        return outputs

    def postprocess(self, model_outputs):
        """
        Every list entry in the output is of the type {verb: (labels, List[(token, label)])}
        """
        result = []
        id2label = {int(k): str(v) for k, v in self.model.config.id2label.items()}
        evaluator = Evaluator(id2label)

        for model_output in model_outputs:
            class_probabilities = model_output["class_probabilities"]
            attention_mask = model_output["attention_mask"]
            output_dict = evaluator.make_output_human_readable(
                class_probabilities, attention_mask
            )
            # Here we always fetch the first list because in a pipeline every sentence is processed one at a time
            wordpiece_label_ids = output_dict["wordpiece_label_ids"][0]
            labels = list(map(lambda idx: id2label[idx], wordpiece_label_ids))
            result.append({model_output["verb"]: (labels, list(zip(model_output["tokens"], labels)))})
        return result

    def _find_word(self, s, start=0):
        for i, char in enumerate(s[start:], start):
            if not char.isalpha():
                return start, i
        return start, len(s)



"""
This is a script to generate visualization of
brat files in a batch way.
"""
import plantuml
import matplotlib.pyplot as plt

import networkx as nx
from netgraph import Graph, ArcDiagram, InteractiveGraph
import warnings

import argparse
import os

import text2story.brat2viz.drs2viz.parser as parser
from text2story.brat2viz.brat2drs import brat2drs

def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")

def list_ann_files(inputdir):
    """
    It collects the names of .ann files in the inputdir directory
    """

    lst_files = []
    for dirpath, dirnames, filenames in os.walk(inputdir):
        for f in filenames:
            if f.endswith(".ann"):
                lst_files.append(os.path.join(dirpath,f))

    return lst_files

def build_drs(file_name_ann, file_name_drs):

    filecontent = brat2drs.read_file(file_name_ann)
    f_parser = brat2drs.file_parser(filecontent)
    dexpr_list, f_parser = brat2drs.assign_variable(f_parser)
    dr_set, dexpr_list = brat2drs.attributes_events(dexpr_list, f_parser)
    dexpr_list = brat2drs.event_event_relation(f_parser, dexpr_list)

    actors, actors_sr = brat2drs.actors_relation(f_parser)

    relations, dexpr_list = brat2drs.TE_relations(f_parser, dexpr_list)
    relations, dexpr_list = brat2drs.TT_relations(relations, f_parser, dexpr_list)

    actors = brat2drs.update_actors(relations, actors, f_parser)

    brat2drs.write_output(dexpr_list, dr_set, actors, relations, file_name_drs)

def create_plantuml(actors_dict, events_dict, events_relations, non_event_relations):

    msc_str = ""
    for x in non_event_relations:
        ref1 = x[0]
        ref2 = x[2]
        rel_type = x[1]
        msc_str = msc_str + "\"%s\"->\"%s\":%s\n" % (actors_dict[ref1], actors_dict[ref2], rel_type)

    for e in events_relations:
        for (ref1,rel_type,ref2) in events_relations[e]:

            # check for redundandt actor inside string and displays only the first one
            actors_ref1_lst = actors_dict[ref1].split(",")
            actors_ref2_lst = actors_dict[ref2].split(",")

            if len(actors_ref1_lst) > 1:
                participant1 = "%s (%d)" % (actors_ref1_lst[0], len(actors_ref1_lst))
            else:
                participant1 = "%s" % (actors_ref1_lst[0])

            if len(actors_ref2_lst) > 1:
                participant2 = "%s (%d)" % (actors_ref2_lst[0], len(actors_ref2_lst))
            else:
                participant2 = "%s" % (actors_ref2_lst[0])

            msc_str = msc_str + "\"%s\"->\"%s\":%s (%s)\n" % (participant1, participant2, events_dict[e],rel_type)
        #print("-->",e,events_relations[e])

    return msc_str

def build_graph(actors_dict, events_dict, events_relations, non_event_relations):

    #G = nx.MultiDiGraph()
    G = nx.DiGraph()

    labels = {}
    key2labels = {}
    for e in events_relations:
        for (ref1,rel_type,ref2) in events_relations[e]:

            # check for redundandt actor inside string and displays only the first one
            actors_ref1_lst = actors_dict[ref1].split(",")
            actors_ref2_lst = actors_dict[ref2].split(",")

            if len(actors_ref1_lst) > 1:
                participant1 = "%s (%d)" % (actors_ref1_lst[0], len(actors_ref1_lst))
            else:
                participant1 = "%s" % (actors_ref1_lst[0])

            if len(actors_ref2_lst) > 1:
                participant2 = "%s (%d)" % (actors_ref2_lst[0], len(actors_ref2_lst))
            else:
                participant2 = "%s" % (actors_ref2_lst[0])

            if participant1 not in key2labels:
                G.add_node(len(key2labels))
                key2labels[participant1] = len(key2labels.keys())

            node1 = key2labels[participant1]

            if participant2 not in key2labels:
                G.add_node(len(key2labels))
                key2labels[participant2] = len(key2labels.keys())

            node2 = key2labels[participant2]

            label_tokens = events_dict[e].split()
            if len(label_tokens) < 4:
                label_edge = "%s" % (events_dict[e])
            else:
                label_edge = " ".join(label_tokens[:3]) + "..."

            G.add_edge(node1, node2, label=label_edge)

            # print("node1: %d node2: %d label_edge:%s" % (node1, node2, label_edge))

    for k in key2labels:
        labels[key2labels[k]] = k

    return G, labels

def process_edge_labels(edge_labels):
    """
    It process edge labels to find if there are more than one label per node, then
    join the labels in one label

    @param edge_labels: a dictionary that contains (tuple):string as a map for edges and their labels.
    @return: Dict
    """
    new_edge_labels = {}

    for edge in edge_labels:
        n0 = edge[0]
        n1 = edge[1]

        if (n0, n1) in new_edge_labels:
            new_edge_labels[(n0, n1)] = new_edge_labels[(n0, n1)] + " |\n " + edge_labels[edge]
        else:
            new_edge_labels[(n0, n1)] = edge_labels[edge]

    return new_edge_labels

def drs2kg(drs_file, output=None):
    #plt.tight_layout()
    #plt.figure(figsize=(9, 9))

    if output is None:
        file_name_out = os.path.basename(drs_file)
        file_name_out = os.path.splitext(drs_file)[0]
    else:
        file_name_out = output

    # get entities
    actors_dict, events_dict, events_relations, non_event_relations = \
            parser.parse_drs(drs_file)

    file_name_graph_fig = "%s_graph.png" % (file_name_out)

    G, labels = build_graph(actors_dict, events_dict, \
                    events_relations, non_event_relations)
    edge_labels = nx.get_edge_attributes(G, "label")

    edge_labels = process_edge_labels(edge_labels)

    node_size = {}
    for nnode in labels:
        if len(labels[nnode]) > 5:
            node_size[nnode] = 7
        else:
            node_size[nnode] = len(labels[nnode])

    fig, ax = plt.subplots(figsize=(9, 9))
    plot_instance = ArcDiagram(G, edge_layout_kwargs=dict(rad=0.5), node_labels=labels,\
                          node_label_fontdict=dict(size=10),\
                          node_size=node_size, edge_labels=edge_labels, node_layout="circular",\
                               edge_label_fontdict=dict(fontsize=9))


    fig.savefig(file_name_graph_fig, format='png')

def drs2kginteractive(drs_file, output=None):
    #plt.tight_layout()
    #plt.figure(figsize=(9, 9))

    if output is None:
        file_name_out = os.path.basename(drs_file)
        file_name_out = os.path.splitext(drs_file)[0]
    else:
        file_name_out = output

    # get entities
    actors_dict, events_dict, events_relations, non_event_relations = \
            parser.parse_drs(drs_file)

    file_name_graph_fig = "%s_graph.png" % (file_name_out)

    G, labels = build_graph(actors_dict, events_dict, \
                    events_relations, non_event_relations)
    edge_labels = nx.get_edge_attributes(G, "label")

    edge_labels = process_edge_labels(edge_labels)

    node_size = {}
    for nnode in labels:
        if len(labels[nnode]) > 5:
            node_size[nnode] = 7
        else:
            node_size[nnode] = len(labels[nnode])


    fig, ax = plt.subplots(figsize=(9, 9))
    plot_instance = InteractiveGraph(G, edge_layout_kwargs=dict(rad=0.5), node_labels=labels,\
                          node_label_fontdict=dict(size=10),\
                          node_size=node_size, edge_labels=edge_labels, node_layout="spring",\
                               edge_label_fontdict=dict(fontsize=6))

    plt.show()

def drs2msc(drs_file, output=None):
    # url server to generate the figures
    url = "http://www.plantuml.com/plantuml/img/"
    pl = plantuml.PlantUML(url)

    if output is None:
        file_name_out = os.path.basename(drs_file)
        file_name_out = os.path.splitext(drs_file)[0]
    else:
        file_name_out = output

    # get entities
    actors_dict, events_dict, events_relations, non_event_relations = \
        parser.parse_drs(drs_file)

    # build plantuml txt file
    msc_str = create_plantuml(actors_dict, events_dict, \
                              events_relations, non_event_relations)

    with open(file_name_out, "w") as fd:
        fd.write(msc_str)

    file_name_fig = "%s.png" % (file_name_out)
    success = pl.processes_file(file_name_out, outfile=file_name_fig)



def drs2vis(drs_file):
    warnings.warn("drs2viz is deprecated, use drs2msc or drs2kg", DeprecationWarning)

    # url server to generate the figures
    url = "http://www.plantuml.com/plantuml/img/"
    pl = plantuml.PlantUML(url)

    file_name_out = os.path.basename(drs_file)
    file_name_out = os.path.splitext(drs_file)[0]

    # get entities
    actors_dict, events_dict, events_relations, non_event_relations = \
            parser.parse_drs(drs_file)

    # build plantuml txt file
    msc_str = create_plantuml(actors_dict, events_dict, \
                                events_relations, non_event_relations)


    with open(file_name_out,"w") as fd:
        fd.write(msc_str)

    file_name_fig = "%s.png" % (file_name_out)

    success = pl.processes_file(file_name_out, outfile=file_name_fig)




def msc_vis(lst_files, outputdir):

    # url server to generate the figures
    url = "http://www.plantuml.com/plantuml/img/"
    pl = plantuml.PlantUML(url)

    for f in lst_files:

        file_name_out = os.path.basename(f)
        file_name_out = os.path.splitext(file_name_out)[0]

        # build drs
        file_name_drs = os.path.join(outputdir, "%s.drs" % (file_name_out))
        build_drs(f, file_name_drs)

        # get entities
        actors_dict, events_dict, events_relations, non_event_relations = \
                parser.parse_drs(file_name_drs)

        # build plantuml txt file
        msc_str = create_plantuml(actors_dict, events_dict, \
                                    events_relations, non_event_relations)
        #print(msc_str)
        file_name_plantuml = os.path.join(outputdir,"%s.txt" % (file_name_out))
        with open(file_name_plantuml,"w") as fd:
            fd.write(msc_str)

        file_name_fig = os.path.join(outputdir, "%s.png" % (file_name_out))

        success = pl.processes_file(file_name_plantuml, outfile=file_name_fig)

#if __name__ == "__main__":

#    my_parser = argparse.ArgumentParser(description='This is a script to generate visualization of\
#brat files in a batch way.')

#    my_parser.add_argument("inputdir", action='store', type=dir_path, help="The directory that \
#            contains the target files (brat format) and the txt narrative files.")
#    my_parser.add_argument("outputdir", action='store', type=dir_path, help="The directory that \
#            will be written the MSC files.")

    # TODO: visualization in graph format

#    args = my_parser.parse_args()

#    ann_files = list_ann_files(args.inputdir)
#    msc_vis(ann_files, args.outputdir)



"""
	text2story.core.link_structures

	Link structure classes (Temporal, aspectual, subordination, semantic role and objectal)
"""

from text2story.core.exceptions import  InvalidLink

class MeasureLink:
    def __init__(self, arg1, arg2, type='distance', id_rel = ""):
        self.type = type
        self.arg1 = arg1
        self.arg2 = arg2
        self.id = id_rel

    def __str__(self):
        return f"{self.id}\tMLINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"


class MovementLink:
    def __init__(self, arg1, arg2, type='spatialRelation', id_rel = ""):
        self.type = type
        self.arg1 = arg1
        self.arg2 = arg2
        self.id = id_rel

    def __str__(self):
        return f"{self.id}\tMOVELINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"


class QualitativeSpatialLink:
    def __init__(self, arg1, arg2, type='figure', id_rel = ""):
        self.type = type
        self.arg1 = arg1
        self.arg2 = arg2
        self.id = id_rel
    def __str__(self):
        return f"{self.id}\tQSLINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"



class SubordinationLink:

    def __init__(self, arg1, arg2, type='factive',id_rel = ""):
        self.type = type
        self.arg1 = arg1
        self.arg2 = arg2
        self.id = id_rel
    def __str__(self):
        return f"{self.id}\tSLINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"



class SemanticRoleLink:
    def __init__(self, actor, event, type="theme", id_rel=""):
        self.type = type
        self.arg1 = actor
        self.arg2 = event
        self.id = id_rel

    def __str__(self):
        return f"{self.id}\tSRLINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"


class ObjectalLink:
    def __init__(self, arg1, arg2, type='objIdentity', id_rel = ""):
        self.type = type
        self.arg1 = arg1
        self.arg2 = arg2
        self.id = id_rel
    def __str__(self):
        return f"{self.id}\tOLINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"


class TemporalLink:
    def __init__(self, arg1, arg2, type='includes', id_rel = ""):
        self.type = type
        self.arg1 = arg1
        self.arg2 = arg2
        self.id = id_rel

    def __str__(self):
        return f"{self.id}\tTLINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"

class AspectualLink:
    def __init__(self, arg1, arg2, type='initiates', id_rel = ""):
        self.type = type
        self.arg1 = arg1
        self.arg2 = arg2
        self.id = id_rel

    def __str__(self):
        return f"{self.id}\tALINK_{self.type} Arg1:{self.arg1} Arg2:{self.arg2}"

def createLinkObject( arg1, arg2, type, subtype, id_rel):
    if type == "TLINK":
        return TemporalLink(arg1, arg2, subtype, id_rel)
    elif type == "OLINK":
        return ObjectalLink(arg1, arg2, subtype, id_rel)
    elif type == "SRLINK":
        return SemanticRoleLink(arg1, arg2, subtype, id_rel)
    elif type == "SLINK":
        return SubordinationLink(arg1, arg2, subtype, id_rel)
    elif type == "QSLINK":
        return QualitativeSpatialLink(arg1, arg2, subtype, id_rel)
    elif type == "MOVELINK":
        return MovementLink(arg1, arg2, subtype, id_rel)
    elif type == "ALINK":
        return AspectualLink(arg1, arg2, subtype, id_rel)
    elif type == "MLINK":
        return MeasureLink(arg1, arg2, subtype, id_rel)
    else:
        raise InvalidLink(type)



from text2story.core.exceptions import InvalidLanguage, UninstalledModel

try:
    import spacy
except ImportError:
    raise  ImportError("To use TEI2GO library, you should install spacy: pip install spacy")

pipeline = {}

def load(lang):
    """
    Used, at start, to load the pipeline for the supported languages.
    """
    if lang not in ["fr", "it", "de", "pt", "es", "en"]:
        raise InvalidLanguage(lang)

    try:
        pipeline[lang] = spacy.load(lang + "_tei2go")
    except OSError:
        model_name = lang + "_tei2go"
        command = f"pip install https://huggingface.co/hugosousa/{lang}_tei2go/resolve/main/{lang}_tei2go-any-py3-none-any.whl"
        raise UninstalledModel(model_name, command)

def extract_times(lang, text, publication_time=None):
    """
    Parameters
    ----------
    lang : str
        the language of text to be annotated

    text : str
        the text to be annotated

    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        a list consisting of the times identified, where each time is represented by a tuple
        with the start and end character offset, it's value and type, respectively

    Raises
    ------
    InvalidLanguage if the language given is invalid/unsupported
    """
    if lang not in ["fr", "it", "de", "pt", "es", "en"]:
        raise InvalidLanguage(lang)

    timex_lst = pipeline[lang](text).ents

    ans = []
    for timex in timex_lst:

        start = timex.start_char
        end = timex.end_char
        label = timex.label_
        text = timex.text

        ans.append(((start, end), label, text))
    return ans


import os
import shlex
from pathlib import Path

import subprocess
import platform
import uuid

import joblib
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))

#from text2story.core.utils import is_library_installed, find_target_dir
from core.utils import is_library_installed, find_target_dir

current_path = Path(__file__).parent
# Path to the virtual environment's activation script
env_name = os.path.join(current_path, "allennlp_venv")
allen_script = os.path.join(current_path, "allen_wrapper.py")

def setup_enviroment():

    # Check if the VIRTUAL_ENV environment variable is set
    if 'VIRTUAL_ENV' in os.environ:
        # Extract the virtual environment name from the path
        env_path = os.environ['VIRTUAL_ENV']
        print(f"Current virtual environment: {os.path.basename(env_path)}")
    else:
        print("No virtual environment is currently activated.")

    # this model requires a older version of transformers library, so
    # this code test it and installed a newer version in a separate directory

    if not(os.path.exists(env_name)):
        # Create a virtual environment with the specified name
        subprocess.call(['python3', '-m', 'venv', env_name])

    # run the setup environment script
    # Determine the appropriate activation script based on the current platform
    if platform.system() == 'Windows':
        activation_script = os.path.join(current_path, "scripts", 'setup_env.bat')
        subprocess.call([activation_script, env_name], shell=True)
    else:
        target_dir = find_target_dir(env_name, "site-packages")

        if not(is_library_installed("allennlp", target_dir)):
            #command = f'source {env_name}/bin/activate && pip install -r {current_path}/requirements.txt'
            command = f'pip install --target={target_dir} -r {current_path}/requirements.txt'
            subprocess.call(command, shell=True,  executable="/bin/bash")


def load(lang):
    # this install the enviroment to run the allen nlp wrapper in
    # the new enviroment
    setup_enviroment()

def extract_participants(lang, text):

    venv_activation_script = os.path.join(env_name, "bin", "activate")

    # Generate a random UUID
    unique_id = str(uuid.uuid4())
    input_file = os.path.join(current_path, f"participant_{unique_id}.txt")
    with open(input_file, "w") as fd:
        fd.write(text)

    output_file = os.path.join(current_path, f"participant_output_{unique_id}.joblib")
    subprocess.call(f"source {venv_activation_script} && python {allen_script} --lang {lang} --action participant \
    --input {input_file} --output {output_file}", shell=True,  executable="/bin/bash")

    result = joblib.load(output_file)

    os.remove(output_file)
    os.remove(input_file)

    return result

def extract_events(lang, text):
    venv_activation_script = os.path.join(env_name, "bin", "activate")

    # Generate a random UUID
    unique_id = str(uuid.uuid4())
    input_file = os.path.join(current_path, f"event_{unique_id}.txt")
    with open(input_file, "w") as fd:
        fd.write(text)

    output_file = os.path.join(current_path, f"event_output_{unique_id}.joblib")
    subprocess.call(f"source {venv_activation_script} && python {allen_script} --lang {lang} --action event \
        --input {input_file} --output {output_file}", shell=True, executable="/bin/bash")

    #result = pd.read_pickle(output_file)
    result = joblib.load(output_file)

    os.remove(output_file)
    os.remove(input_file)

    return result

def extract_semantic_role_links(lang, text):

    venv_activation_script = os.path.join(env_name, "bin", "activate")

    # Generate a random UUID
    unique_id = str(uuid.uuid4())
    input_file = os.path.join(current_path, f"srlink_{unique_id}.txt")
    with open(input_file, "w") as fd:
        fd.write(text)

    output_file = os.path.join(current_path, f"srlink_output_{unique_id}.joblib")
    subprocess.call(f"source {venv_activation_script} && python {allen_script} --lang {lang} --action srlink \
            --input {input_file} --output {output_file}", shell=True, executable="/bin/bash")

    # result = pd.read_pickle(output_file)
    result = joblib.load(output_file)

    os.remove(output_file)
    os.remove(input_file)

    return result

def extract_objectal_links(lang, text):

    venv_activation_script = os.path.join(env_name, "bin", "activate")

    # Generate a random UUID
    unique_id = str(uuid.uuid4())
    working_dir = os.getcwd()
    input_file = os.path.join(working_dir, f"olink_{unique_id}.txt")
    with open(input_file, "w") as fd:
        fd.write(text)

    output_file = os.path.join(working_dir, f"olink_output_{unique_id}.joblib")
    # print(f"source {venv_activation_script} && python {allen_script} --lang {lang} --action olink \
    #            --input {input_file} --output {output_file}")

    subprocess.call(f"source {venv_activation_script} && python {allen_script} --lang {lang} --action olink \
            --input {input_file} --output {output_file}", shell=True, executable="/bin/bash")

    # result = pd.read_pickle(output_file)
    result = joblib.load(output_file)

    os.remove(output_file)
    os.remove(input_file)

    return result

def extract_srl(lang, text):
    venv_activation_script = os.path.join(env_name, "bin", "activate")

    # Generate a random UUID
    unique_id = str(uuid.uuid4())
    working_dir = os.getcwd()
    input_file = os.path.join(working_dir, f"srl_{unique_id}.txt")
    with open(input_file, "w") as fd:
        fd.write(text)

    output_file = os.path.join(working_dir, f"srl_output_{unique_id}.joblib")
    # print(f"source {venv_activation_script} && python {allen_script} --lang {lang} --action olink \
    #            --input {input_file} --output {output_file}")

    subprocess.call(f"source {venv_activation_script} && python {allen_script} --lang {lang} --action srl \
                --input {input_file} --output {output_file}", shell=True, executable="/bin/bash")

    # result = pd.read_pickle(output_file)
    result = joblib.load(output_file)

    os.remove(output_file)
    os.remove(input_file)

    return result





from maverick import Maverick
from text2story.core.exceptions import InvalidLanguage

pipeline = {}
def load(lang):
    if lang == "en":
        pipeline['coref_en'] = Maverick(hf_name_or_path="sapienzanlp/maverick-mes-ontonotes", device="cpu")
    else:
        raise InvalidLanguage(lang)

def extract_objectal_links(lang, text):
    prediction = pipeline['coref_en'].predict(text)

    cluster_indexes_list = prediction["clusters_char_offsets"]
    # for some reason, maverick model is cutting the last character
    #, so in the following line we update the end

    cluster_indexes_list = [[ (start, end + 1) for (start, end) in cluster]   for cluster in cluster_indexes_list]

    return cluster_indexes_list




"""
    text2story/core/utils module

    Functions
    ---------
    chunknize_actors(annotations)
        converts the result from the 'extract_actors' method, implemented by the annotators supported, to a list the chunknized list of actors
        to do this conversion, the IOB in the NE tag is used
"""

from itertools import tee
import subprocess
import sys

import os

import warnings

import importlib_metadata
import numpy as np
from typing import Tuple, List


def bsearch_tuplelist(x: int, xs: List[Tuple[int, int]]) -> int:
    """
    Use binary search to check if x is whithin some interval of a tuple list
    xs.

    if x is not in none of intervals, it return -1, otherwise returns de
    position of the interval.
    @type xs: object

    """
    b = 0
    e = len(xs) - 1
    m = (b + e) // 2
    pos = -1

    while b <= e:
        (x_b, x_e) = xs[m]

        if x>= x_b and x<= x_e:
            return m
        elif x < x_b:
            e = m - 1
        elif x > x_e:
            b = m + 1

        m = (b + e) // 2

    return pos



def join_tokens(tok_lst):

    special_suffix = ["se","me", "te", "lhe", "o", "a", "la", "lo", "lho", "lha", "nos","feira"]
    parentheses = ["(", "{"]
    punctuation = [",", ".", ":", ";", "?", "!", ")", "}"]

    # first, find all hyphen in the token list
    array_tok_lst = np.array(tok_lst)
    indices = np.where(array_tok_lst == "-")[0]
    indices = list(indices)

    if indices != []:
        tok_join = ""
        if len(tok_lst) > 0:
            i = 0
            while i < len(tok_lst):
                if i in indices:
                    if i + 1 < len(tok_lst) and \
                            tok_lst[i + 1].lower() in special_suffix:
                        tok_join = tok_join.strip() + tok_lst[i]
                        i = i + 1

                if tok_lst[i] in parentheses:
                    tok_join = tok_join + tok_lst[i]
                else:
                    if tok_lst[i] in punctuation:
                        tok_join = tok_join.rstrip() + tok_lst[i] + " "
                    else:
                        tok_join = tok_join + tok_lst[i] + " "
                i = i + 1

        return tok_join.strip()
    else:
        tok_join = ""
        if len(tok_lst) > 0:
            i = 0
            while i < len(tok_lst):

                if tok_lst[i] in parentheses:
                    tok_join = tok_join + tok_lst[i]
                else:
                    if tok_lst[i] in punctuation:
                        tok_join = tok_join.rstrip() + tok_lst[i] + " "
                    else:
                        tok_join = tok_join + tok_lst[i] + " "
                i = i + 1

        return tok_join.strip()


def chunknize_actors(annotations):
    """
    Parameters
    ----------
    annotations : list[tuple[tuple[int, int], str, str]]
        list of annotations made by some tool for each token

    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of actors identified where each actor is represented by a tuple
    """

    actors = []

    ready_to_add = False

    prev_ne_tag = ''

    for ann in annotations:
        token_character_span, token_pos_tag, token_ne_iob_tag = ann

        # token_ne_iob_tag = 'I-PER', then token_ne_iob_tag[2:] == 'PER'
        if token_ne_iob_tag.startswith("B") or (
                token_ne_iob_tag.startswith("I") and token_ne_iob_tag[2:] != prev_ne_tag):
            # Case we start a new chunk, after finishing another, for instance: Case B-Per, I-Per, B(or I)-Org, then we add the finished actor
            if ready_to_add:
                actor = ((actor_start_offset, actor_end_offset), actor_lexical_head, actor_actor_type)
                actors.append(actor)

            ready_to_add = True
            actor_start_offset = token_character_span[0]
            actor_end_offset = token_character_span[1]
            actor_lexical_head = token_pos_tag if token_pos_tag in ['Noun', 'Pronoun'] else 'UNDEF'
            actor_actor_type = token_ne_iob_tag[2:]

        elif token_ne_iob_tag.startswith("I"):
            # actor_start_offset it's always the same, since it's defined by the first token of the actor
            actor_end_offset = token_character_span[1]
            actor_lexical_head = actor_lexical_head if actor_lexical_head != 'UNDEF' else token_pos_tag if token_pos_tag in [
                'Noun', 'Pronoun'] else 'UNDEF'
            # actor_actor_type it's the same for all tokens that constitute the actor and it's already defined by the first token of the actor

        elif token_ne_iob_tag.startswith("O") and ready_to_add:
            actor = ((actor_start_offset, actor_end_offset), actor_lexical_head, actor_actor_type)
            actors.append(actor)
            ready_to_add = False
            # No need to reset the variables, since the first update writes over

        prev_ne_tag = token_ne_iob_tag[2:]

    if ready_to_add:  # If the last token still makes part of the actor
        actor = ((actor_start_offset, actor_end_offset), actor_lexical_head, actor_actor_type)
        actors.append(actor)

    return actors


def pairwise(iterable):
    """
    Iterate through some iterable with a lookahead.
    From the itertools docs recipes - https://docs.python.org/3/library/itertools.html
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)

def merge_dicts(d1, d2):
    for k2 in d2:
        for attr in d2[k2]:
            if k2 in d1:
                if attr in d1[k2]:
                    d1[k2][attr] += d2[k2][attr]
                else:
                    d1[k2][attr] = d2[k2][attr]

            else:
                d1[k2] = {attr:d2[k2][attr]}

def capfirst(s):
    return s[:1].upper() + s[1:]

def reset_offset_ann(doc, txt, start_offset=0, start_txt=0):
    """
    Given a document in tokencorpus format, looks for the spans in txt
    and then reset the span
    @param doc: A TokenCorpus LIst
    @param txt: a string
    @return: a token corpus list with new spans
    """
    start = start_txt
    for tok in doc:
        if tok.offset >= start_offset:
            start_tok = txt.find(tok.text, start)
            if start_tok == -1:
                warnings.warn("reset_offset_ann: The token %s was not found in given text." % tok.text)
            else:
                #print(">>", tok.text, tok.offset, start_tok)
                tok.offset = start_tok
                start = start_tok + len(tok.text)

    return doc

def diff_ann(doc1, doc2):
    ann1 = {}
    for tok in doc1:
        if tok.attr is not None:
            for id in tok.id_ann:
                if id in ann1:
                    ann1[id] = ann1[id] + " " + tok.text
                else:
                    ann1[id] = tok.text
    ann2 = {}
    for tok in doc2:
        if tok.attr is not None:
            for id in tok.id_ann:
                if id in ann2:
                    ann2[id] = ann2[id] + " " + tok.text
                else:
                    ann2[id] = tok.text


def normalize_tag(label):
    """
    Parameters
    ----------
    label : str

    Returns
    -------
    str
        the label normalized
    """

    mapping = {
        # POS tags
        # Universal POS Tags
        # http://universaldependencies.org/u/pos/

        # "ADJ": "adjective",
        # "ADP": "adposition",
        # "ADV": "adverb",
        # "AUX": "auxiliary",
        # "CONJ": "conjunction",
        # "CCONJ": "coordinating conjunction",
        # "DET": "determiner",
        # "INTJ": "interjection",
        "NOUN": "Noun",
        # "NUM": "numeral",
        # "PART": "particle",
        "PRON": "Pronoun",
        "PROPN": "Noun",
        "NPROP":"Noun",
        # "PUNCT": "punctuation",
        # "SCONJ": "subordinating conjunction",
        # "SYM": "symbol",
        # "VERB": "verb",
        # "X": "other",
        # "EOL": "end of line",
        # "SPACE": "space",

        # NE
        # en
        'CARDINAL': 'Other',  # 'Numerals that do not fall under another type'
        'DATE': 'Date',  # 'Absolute or relative dates or periods'
        'EVENT': 'Other',  # 'Named hurricanes, battles, wars, sports events, etc.'
        'FAC': 'Loc',  # 'Buildings, airports, highways, bridges, etc.'
        'GPE': 'Loc',  # 'Countries, cities, states'
        'LANGUAGE': 'Other',  # 'Any named language'
        'LAW': 'Other',  # 'Named documents made into laws.'
        'LOC': 'Loc',  # 'Non-GPE locations, mountain ranges, bodies of water'
        'MONEY': 'Other',  # 'Monetary values, including unit'
        'NORP': 'Other',  # 'Nationalities or religious or political groups'
        'ORDINAL': 'Other',  # '"first", "second", etc.'
        'ORG': 'Org',  # 'Companies, agencies, institutions, etc.'
        'PERCENT': 'Other',  # 'Percentage, including "%"'
        'PERSON': 'Per',  # 'People, including fictional'
        'PRODUCT': 'Obj',  # 'Objects, vehicles, foods, etc. (not services)'
        'QUANTITY': 'Other',  # 'Measurements, as of weight or distance'
        'TIME': 'Time',  # 'Times smaller than a day'
        'WORK_OF_ART': 'Other',  # 'Titles of books, songs, etc.'

        # pt
        # 'LOC'
        'MISC': 'Other',  # 'Miscellaneous entities, e.g. events, nationalities, products or works of art'
        # 'ORG'
        'PER': 'Per'  # 'People, including fictional'
    }

    return mapping.get(label, 'UNDEF')


def install(pkg,path):
    return subprocess.check_call([sys.executable, "-m", "pip", "install", pkg,"--target={}".format(path)])


def is_library_installed(library_name, target_directory):
    try:
        installed_distributions = importlib_metadata.distributions(path=[target_directory])

        for distribution in installed_distributions:
            if distribution.metadata['Name'] == library_name:
                return True

        return False
    except Exception:
        return False

def fast_scandir(dirname):
    subfolders= [f.path for f in os.scandir(dirname) if f.is_dir()]
    for dirname in list(subfolders):
        subfolders.extend(fast_scandir(dirname))
    return subfolders

def find_target_dir(pathname, target):
    subfolders = fast_scandir(pathname)
    for p in subfolders:
        if p.endswith(target):
            return p

def map_pos2head(pos_tag):
    p = pos_tag.lower()

    map = {"noun":"Noun","propn":"Noun","pron":"Pronoun","det":"Noun"}
    if p in map:
        return map[p]
    else:
        return "None"

def find_first_non_space(s, p):
    # Check if p is out of range
    if p < 0 or p >= len(s):
        return -1  # Invalid position

    # Start from position p and iterate until a non-space character is found
    while p < len(s) and s[p].isspace():
        p += 1

    # Check if we reached the end of the string without finding a non-space character
    if p >= len(s):
        return -1  # No non-space character found

    return p  # Return the position of the first non-space character

def find_substring_match(matches, position):

    for match in matches:
        # Get the matched substring
        matched_substring = match.group(0)

        # Get the start and end positions of the matched substring
        start_pos = match.start()
        end_pos = match.end()

        # Check if the matched substring is within the desired range (starting from x)
        if start_pos >= position:
            return start_pos

def update_offsets(old_text, new_text, annotation_file):
    """
    Update character offsets in annotation file based on new text positions.

    Args:
        old_text (str): Original text content
        new_text (str): Modified text content
        annotation_file (str): Content of the annotation file

    Returns:
        str: Updated annotation file content
    """
    # Read annotations line by line
    updated_annotations = []
    annotations = annotation_file.strip().split('\n')

    # Create a mapping of old expressions to their new positions
    expression_mapping = {}

    for line in annotations:
        parts = line.split()
        if len(parts) >= 4 and parts[0].startswith('T'):  # Only process T lines
            # Extract the expression from old text using offsets
            start, end = int(parts[2]), int(parts[3])
            expression = old_text[start:end].strip()

            # Find new position of expression in new text
            # Start searching from beginning to maintain order
            search_start = 0
            while True:
                new_start = new_text.find(expression, search_start)
                if new_start == -1:
                    break

                # Check if this position maintains relative ordering
                valid_position = True
                for old_expr, (prev_start, _) in expression_mapping.items():
                    old_pos = old_text.find(old_expr)
                    if (old_pos < start and new_start < prev_start) or \
                            (old_pos > start and new_start > prev_start):
                        continue
                    valid_position = False
                    break

                if valid_position:
                    new_end = new_start + len(expression)
                    expression_mapping[expression] = (new_start, new_end)
                    break
                search_start = new_start + 1

    # Update annotations with new offsets
    for line in annotations:
        parts = line.split()
        if len(parts) >= 4 and parts[0].startswith('T'):
            # Extract the expression from old text
            start, end = int(parts[2]), int(parts[3])
            expression = old_text[start:end].strip()

            if expression in expression_mapping:
                new_start, new_end = expression_mapping[expression]
                # Replace old offsets with new ones
                parts[2] = str(new_start)
                parts[3] = str(new_end)

                updated_line = parts[0] + '\t' + ' '.join(parts[1:4]) + '\t' + ' '.join(parts[4:])
            else:
                updated_line = line  # Keep original if expression not found
        else:
            updated_line = line  # Keep non-T lines unchanged

        updated_annotations.append(updated_line)

    return '\n'.join(updated_annotations)


from typing import Optional, List, Dict
import math
import torch
import evaluate
import logging
from transformers import EvalPrediction


class Evaluator:
    def __init__(self, index2label: Dict[int, str]):
        self.index2label = index2label

    def get_lengths_from_binary_sequence_mask(
        self, mask: torch.BoolTensor
    ) -> torch.LongTensor:
        """
        Compute sequence lengths for each batch element in a tensor using a
        binary mask.

        # Parameters

        mask : `torch.BoolTensor`, required.
            A 2D binary mask of shape (batch_size, sequence_length) to
            calculate the per-batch sequence lengths from.

        # Returns

        `torch.LongTensor`
            A torch.LongTensor of shape (batch_size,) representing the lengths
            of the sequences in the batch.
        """
        # Removes -100 entries which correspond to HF added padding
        return (mask >= 0).sum(-1)

    def viterbi_decode(
        self,
        tag_sequence: torch.Tensor,
        transition_matrix: torch.Tensor,
        tag_observations: Optional[List[int]] = None,
        allowed_start_transitions: torch.Tensor = None,
        allowed_end_transitions: torch.Tensor = None,
        top_k: int = None,
    ):
        """
        Perform Viterbi decoding in log space over a sequence given a transition matrix
        specifying pairwise (transition) potentials between tags and a matrix of shape
        (sequence_length, num_tags) specifying unary potentials for possible tags per
        timestep.

        # Parameters

        tag_sequence : torch.Tensor, required.
            A tensor of shape (sequence_length, num_tags) representing scores for
            a set of tags over a given sequence.
        transition_matrix : torch.Tensor, required.
            A tensor of shape (num_tags, num_tags) representing the binary potentials
            for transitioning between a given pair of tags.
        tag_observations : Optional[List[int]], optional, (default = None)
            A list of length `sequence_length` containing the class ids of observed
            elements in the sequence, with unobserved elements being set to -1. Note that
            it is possible to provide evidence which results in degenerate labelings if
            the sequences of tags you provide as evidence cannot transition between each
            other, or those transitions are extremely unlikely. In this situation we log a
            warning, but the responsibility for providing self-consistent evidence ultimately
            lies with the user.
        allowed_start_transitions : torch.Tensor, optional, (default = None)
            An optional tensor of shape (num_tags,) describing which tags the START token
            may transition *to*. If provided, additional transition constraints will be used for
            determining the start element of the sequence.
        allowed_end_transitions : torch.Tensor, optional, (default = None)
            An optional tensor of shape (num_tags,) describing which tags may transition *to* the
            end tag. If provided, additional transition constraints will be used for determining
            the end element of the sequence.
        top_k : int, optional, (default = None)
            Optional integer specifying how many of the top paths to return. For top_k>=1, returns
            a tuple of two lists: top_k_paths, top_k_scores, For top_k==None, returns a flattened
            tuple with just the top path and its score (not in lists, for backwards compatibility).

        # Returns

        viterbi_path : `List[int]`
            The tag indices of the maximum likelihood tag sequence.
        viterbi_score : `torch.Tensor`
            The score of the viterbi path.
        """
        if top_k is None:
            top_k = 1
            flatten_output = True
        elif top_k >= 1:
            flatten_output = False
        else:
            raise ValueError(
                f"top_k must be either None or an integer >=1. Instead received {top_k}"
            )

        sequence_length, num_tags = list(tag_sequence.size())

        has_start_end_restrictions = (
            allowed_end_transitions is not None or allowed_start_transitions is not None
        )

        if has_start_end_restrictions:

            if allowed_end_transitions is None:
                allowed_end_transitions = torch.zeros(num_tags)
            if allowed_start_transitions is None:
                allowed_start_transitions = torch.zeros(num_tags)

            num_tags = num_tags + 2
            new_transition_matrix = torch.zeros(num_tags, num_tags)
            new_transition_matrix[:-2, :-2] = transition_matrix

            # Start and end transitions are fully defined, but cannot transition between each other.

            allowed_start_transitions = torch.cat(
                [allowed_start_transitions, torch.tensor([-math.inf, -math.inf])]
            )
            allowed_end_transitions = torch.cat(
                [allowed_end_transitions, torch.tensor([-math.inf, -math.inf])]
            )

            # First define how we may transition FROM the start and end tags.
            new_transition_matrix[-2, :] = allowed_start_transitions
            # We cannot transition from the end tag to any tag.
            new_transition_matrix[-1, :] = -math.inf

            new_transition_matrix[:, -1] = allowed_end_transitions
            # We cannot transition to the start tag from any tag.
            new_transition_matrix[:, -2] = -math.inf

            transition_matrix = new_transition_matrix

        if tag_observations:
            if len(tag_observations) != sequence_length:
                raise Exception(
                    f"Observations were provided, but they were not the same length as the sequence. Found sequence of length: {sequence_length} and evidence: {tag_observations}"
                )
        else:
            tag_observations = [-1 for _ in range(sequence_length)]

        if has_start_end_restrictions:
            tag_observations = [num_tags - 2] + tag_observations + [num_tags - 1]
            zero_sentinel = torch.zeros(1, num_tags)
            extra_tags_sentinel = torch.ones(sequence_length, 2) * -math.inf
            tag_sequence = torch.cat([tag_sequence, extra_tags_sentinel], -1)
            tag_sequence = torch.cat([zero_sentinel, tag_sequence, zero_sentinel], 0)
            sequence_length = tag_sequence.size(0)

        path_scores = []
        path_indices = []

        if tag_observations[0] != -1:
            one_hot = torch.zeros(num_tags)
            one_hot[tag_observations[0]] = 100000.0
            path_scores.append(one_hot.unsqueeze(0))
        else:
            path_scores.append(tag_sequence[0, :].unsqueeze(0))

        # Evaluate the scores for all possible paths.
        for timestep in range(1, sequence_length):
            # Add pairwise potentials to current scores.
            summed_potentials = (
                path_scores[timestep - 1].unsqueeze(2) + transition_matrix
            )
            summed_potentials = summed_potentials.view(-1, num_tags)

            # Best pairwise potential path score from the previous timestep.
            max_k = min(summed_potentials.size()[0], top_k)
            scores, paths = torch.topk(summed_potentials, k=max_k, dim=0)

            # If we have an observation for this timestep, use it
            # instead of the distribution over tags.
            observation = tag_observations[timestep]
            # Warn the user if they have passed
            # invalid/extremely unlikely evidence.
            if tag_observations[timestep - 1] != -1 and observation != -1:
                if (
                    transition_matrix[tag_observations[timestep - 1], observation]
                    < -10000
                ):
                    logging.info(
                        "The pairwise potential between tags you have passed as "
                        "observations is extremely unlikely. Double check your evidence "
                        "or transition potentials!"
                    )
            if observation != -1:
                one_hot = torch.zeros(num_tags)
                one_hot[observation] = 100000.0
                path_scores.append(one_hot.unsqueeze(0))
            else:
                path_scores.append(tag_sequence[timestep, :] + scores)
            path_indices.append(paths.squeeze())

        # Construct the most likely sequence backwards.
        path_scores_v = path_scores[-1].view(-1)
        max_k = min(path_scores_v.size()[0], top_k)
        viterbi_scores, best_paths = torch.topk(path_scores_v, k=max_k, dim=0)
        viterbi_paths = []
        for i in range(max_k):
            viterbi_path = [best_paths[i]]
            for backward_timestep in reversed(path_indices):
                viterbi_path.append(int(backward_timestep.view(-1)[viterbi_path[-1]]))
            # Reverse the backward path.
            viterbi_path.reverse()

            if has_start_end_restrictions:
                viterbi_path = viterbi_path[1:-1]

            # Viterbi paths uses (num_tags * n_permutations) nodes; therefore, we need to modulo.
            viterbi_path = [j % num_tags for j in viterbi_path]
            viterbi_paths.append(viterbi_path)

        if flatten_output:
            return viterbi_paths[0], viterbi_scores[0]

        return viterbi_paths, viterbi_scores

    def make_output_human_readable(
        self, class_probabilities, attention_mask
    ) -> Dict[str, torch.Tensor]:
        """
        Does constrained viterbi decoding on class probabilities output in :func:`forward`.  The
        constraint simply specifies that the output tags must be a valid BIO sequence.  We add a
        `"tags"` key to the dictionary with the result.

        NOTE: First, we decode a BIO sequence on top of the wordpieces. This is important; viterbi
        decoding produces low quality output if you decode on top of word representations directly,
        because the model gets confused by the 'missing' positions (which is sensible as it is trained
        to perform tagging on wordpieces, not words).

        Secondly, it's important that the indices we use to recover words from the wordpieces are the
        start_offsets (i.e offsets which correspond to using the first wordpiece of words which are
        tokenized into multiple wordpieces) as otherwise, we might get an ill-formed BIO sequence
        when we select out the word tags from the wordpiece tags. This happens in the case that a word
        is split into multiple word pieces, and then we take the last tag of the word, which might
        correspond to, e.g, I-V, which would not be allowed as it is not preceeded by a B tag.
        """

        sequence_lengths = self.get_lengths_from_binary_sequence_mask(
            attention_mask
        ).data.tolist()

        # Multiple sequences
        if class_probabilities.dim() == 3:
            predictions_list = [
                class_probabilities[i].detach().cpu()
                for i in range(class_probabilities.size(0))
            ]
        else:
            predictions_list = [class_probabilities]

        wordpiece_tags = []
        wordpiece_label_ids = []
        # word_tags = []
        transition_matrix = self.get_viterbi_pairwise_potentials()
        start_transitions = self.get_start_transitions()
        # **************** Different ********************
        # We add in the offsets here so we can compute the un-wordpieced tags.

        output_dict = {}

        for predictions, length in zip(
            predictions_list, sequence_lengths  # , output_dict["start_offsets"]
        ):

            max_likelihood_sequence, _ = self.viterbi_decode(
                predictions[:length],
                transition_matrix,
                allowed_start_transitions=start_transitions,
            )

            tags = [self.index2label[x] for x in max_likelihood_sequence]

            wordpiece_tags.append(tags)
            wordpiece_label_ids.append(max_likelihood_sequence)
            # word_tags.append([tags[i] for i in offsets])

        output_dict["wordpiece_tags"] = wordpiece_tags
        output_dict["wordpiece_label_ids"] = wordpiece_label_ids
        # output_dict["tags"] = word_tags
        return output_dict

    def get_viterbi_pairwise_potentials(self):
        """
        Generate a matrix of pairwise transition potentials for the BIO labels.
        The only constraint implemented here is that I-XXX labels must be preceded
        by either an identical I-XXX tag or a B-XXX tag. In order to achieve this
        constraint, pairs of labels which do not satisfy this constraint have a
        pairwise potential of -inf.

        # Returns

        transition_matrix : torch.Tensor
            A (num_labels, num_labels) matrix of pairwise potentials.
        """
        all_labels = (self.index2label).copy()

        # Workaround: During loss calculation and hf's padding process, -100 is used as the ignore index
        # However, here fo the concept of the transition matrix that would not be applicable, as it would create entries
        # in the -100th row and column. So I temporarily map it to 0 and remmap it back after the decoding
        # ignore_label = all_labels.pop(-100)
        # all_labels[0] = ignore_label

        num_labels = len(all_labels)
        transition_matrix = torch.zeros([num_labels, num_labels])

        for i, previous_label in all_labels.items():
            for j, label in all_labels.items():
                # I labels can only be preceded by themselves or
                # their corresponding B tag.
                if i != j and label[0] == "I" and not previous_label == "B" + label[1:]:
                    transition_matrix[i, j] = float("-inf")
        return transition_matrix

    def get_start_transitions(self):
        """
        In the BIO sequence, we cannot start the sequence with an I-XXX tag.
        This transition sequence is passed to viterbi_decode to specify this constraint.

        # Returns

        start_transitions : torch.Tensor
            The pairwise potentials between a START token and
            the first token of the sequence.
        """
        all_labels = self.index2label
        num_labels = len(all_labels)

        start_transitions = torch.zeros(num_labels)

        for i, label in all_labels.items():
            if label[0] == "I":
                start_transitions[i] = float("-inf")

        return start_transitions


def prepare_compute_metrics(evaluator: Evaluator):
    precision = evaluate.load("precision")
    recall = evaluate.load("recall")
    f1 = evaluate.load("f1")

    id2label = evaluator.index2label

    labels = [
        k for k, v in evaluator.index2label.items() if v.split("-")[-1] != "V"
    ]  # Ignore Bverb labels as that will be provided to the model

    def compute_metrics(eval_prediction: EvalPrediction):
        class_probabilities = torch.tensor(
            eval_prediction.predictions[1]
        )  # Get class probabilities
        attention_mask = torch.tensor(eval_prediction.predictions[2])  # Get mask
        # input_ids = torch.tensor(eval_prediction.predictions[3])

        output_dict = evaluator.make_output_human_readable(
            class_probabilities, attention_mask
        )
        references = eval_prediction.label_ids

        overall_precision = 0
        overall_recall = 0
        overall_f1 = 0

        for i, curr_ref in enumerate(references):
            curr_ref = curr_ref[curr_ref != -100]  # Remove HF padding
            curr_ref = curr_ref.tolist()

            predictions = output_dict["wordpiece_label_ids"][i]

            overall_precision += precision.compute(
                references=curr_ref,
                predictions=predictions,
                labels=labels,
                average="weighted",
                zero_division=0,
            )["precision"]

            overall_recall += recall.compute(
                references=curr_ref,
                predictions=predictions,
                labels=labels,
                average="weighted",
                zero_division=0,
            )["recall"]

            overall_f1 += f1.compute(
                references=curr_ref,
                predictions=predictions,
                labels=labels,
                average="weighted",
            )["f1"]

            print(set(id2label[ref] for ref in curr_ref) - set(id2label[pred] for pred in predictions))

        return {
            "precision": overall_precision / len(references),
            "recall": overall_recall / len(references),
            "f1": overall_f1 / len(references),
        }

    return compute_metrics




from typing import List, Dict, Tuple

from text2story.readers.token_corpus import TokenCorpus
from text2story.core.utils import join_tokens

def select_sentences(doc:List[TokenCorpus]) -> List[Tuple[int,List[TokenCorpus]]]:
    """
    select sentences from a document, i.e., a token corpus list
    @param doc: a token corpus list that represents a text document
    @return: return a list of tuples that maps each sentence id to its correspondent list of tokens
    (each one as a TokenCorpus object)
    """

    sentences = []
    current_sent_id = None
    current_sent = []

    for tok in doc:
        if current_sent_id != None and tok.sent_id == current_sent_id:
            current_sent.append(tok)
        else:
            if current_sent_id is None:
                current_sent.append(tok)
            else:
                sentences.append((current_sent_id, current_sent))
                current_sent = [tok]
            current_sent_id = tok.sent_id

    sentences.append((current_sent_id, current_sent))


    return sentences

def select_sentences_ann(sent_lst:List[Tuple[int,List[TokenCorpus]]], narrative_component:str,**kwargs)->List[Tuple[int,List[TokenCorpus]]]:
    """

    @param sent_lst: A sentence list (the output of the method select_sentences)
    @param narrative_component: 'event','participant', or 'time'
    @param kwargs: attribute list related to the specified component to filter annotations
    @return: a list of sentences with the specified annotations
    """

    #print(len(kwargs))
    sent_id_lst = set()
    for sent_id, sent in sent_lst:
        for tok in sent:

            if tok.attr != []:
                for ann in tok.attr:
                    ann_type, ann_attr = ann
                    if ann_type.lower() == narrative_component:
                        if len(kwargs) > 0:
                            requirements = True
                            for k in kwargs:
                                if k not in ann_attr:
                                    requirements = False
                                    break
                                else:
                                    if ann_attr[k] != kwargs[k]:
                                        requirements = False
                                        break
                            if requirements:
                                sent_id_lst.add(sent_id)
                        else:
                            sent_id_lst.add(sent_id)
    ans = []

    for (sent_id, sent) in sent_lst:
        if sent_id in sent_id_lst:
            ans.append((sent_id, sent))

    return ans



import logging
from typing import Dict, List, Iterable, Tuple, Any

from overrides import overrides
from transformers import AutoTokenizer

from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import Field, TextField, SequenceLabelField, MetadataField
from allennlp.data.instance import Instance
from allennlp.data.token_indexers import SingleIdTokenIndexer, TokenIndexer
from allennlp.data.tokenizers import Token, Tokenizer

#from text2story.annotators.CUSTOMPT.srl_bert_pt.PropBankBr import PropBankBr, PropBankBrSentence
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../../')))

from ALLENNLP.PropBankBr import PropBankBr, PropBankBrSentence

from allennlp_models.structured_prediction.dataset_readers import UniversalDependenciesDatasetReader
from conllu import parse_incr

#import text2story.annotators.CUSTOMPT.srl_bert_pt.preprocess
#import annotators.ALLENNLP.preprocess
import preprocess
import codecs

# This file is heavily based on the dataset readers from allennlp_models
# But the original file couldn't be used with xlmr models (at the time)


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _convert_tags_to_wordpiece_tags(tags: List[str], offsets: List[int]) -> List[str]:
    """
    Converts a series of BIO tags to account for a wordpiece tokenizer,
    extending/modifying BIO tags where appropriate to deal with words which
    are split into multiple wordpieces by the tokenizer.

    This is only used if you pass a `bert_model_name` to the dataset reader below.

    # Parameters

    tags : `List[str]`
        The BIO formatted tags to convert to BIO tags for wordpieces
    offsets : `List[int]`
        The wordpiece offsets.

    # Returns

    The new BIO tags.
    """
    new_tags = []
    j = 0
    for i, offset in enumerate(offsets):
        tag = tags[i]
        is_o = tag == "O"
        is_start = True
        while j < offset:
            if is_o:
                new_tags.append("O")

            elif tag.startswith("I"):
                new_tags.append(tag)

            elif is_start and tag.startswith("B"):
                new_tags.append(tag)
                is_start = False

            elif tag.startswith("B"):
                _, label = tag.split("-", 1)
                new_tags.append("I-" + label)
            j += 1

    # Add O tags for cls and sep tokens.
    return ["O"] + new_tags + ["O"]


def _convert_verb_indices_to_wordpiece_indices(verb_indices: List[int], offsets: List[int]):
    """
    Converts binary verb indicators to account for a wordpiece tokenizer,
    extending/modifying BIO tags where appropriate to deal with words which
    are split into multiple wordpieces by the tokenizer.

    This is only used if you pass a `bert_model_name` to the dataset reader below.

    # Parameters

    verb_indices : `List[int]`
        The binary verb indicators, 0 for not a verb, 1 for verb.
    offsets : `List[int]`
        The wordpiece offsets.

    # Returns

    The new verb indices.
    """
    j = 0
    new_verb_indices = []
    for i, offset in enumerate(offsets):
        indicator = verb_indices[i]
        while j < offset:
            new_verb_indices.append(indicator)
            j += 1

    # Add 0 indicators for cls and sep tokens.
    return [0] + new_verb_indices + [0]



@DatasetReader.register("srl-pt")
class SrlReader(DatasetReader):
    """
    This dataset reader is almost identical to allennlp_models/syntax/srl/srl_reader.py
    But the original file couldn't be used with xlmr models (at the time)
    And the _read was changed to read folds data.

    Parameters
    ----------
    token_indexers : ``Dict[str, TokenIndexer]``, optional
        We similarly use this for both the premise and the hypothesis.  See :class:`TokenIndexer`.
        Default is ``{"tokens": SingleIdTokenIndexer()}``.
    domain_identifier: ``str``, (default = None)
        A string denoting a sub-domain of the Ontonotes 5.0 dataset to use. If present, only
        conll files under paths containing this domain identifier will be processed.
    bert_model_name : ``Optional[str]``, (default = None)
        The BERT model to be wrapped. If you specify a bert_model here, then we will
        assume you want to use BERT throughout; we will use the bert tokenizer,
        and will expand your tags and verb indicators accordingly. If not,
        the tokens will be indexed as normal with the token_indexers.

    Returns
    -------
    A ``Dataset`` of ``Instances`` for Semantic Role Labelling.
    """
    def __init__(self,
                 token_indexers: Dict[str, TokenIndexer] = None,
                 domain_identifier: str = None,
                 lazy: bool = False,
                 bert_model_name: str = None) -> None:
        super().__init__(lazy)
        self._token_indexers = token_indexers or {"tokens": SingleIdTokenIndexer()}
        self._domain_identifier = domain_identifier
        self.xlm = "xlm" in bert_model_name
        
        self.bert_tokenizer = AutoTokenizer.from_pretrained(bert_model_name, use_fast = False)

        # the model class is different in xlmr models 
        # the vocab is not in the same place as in bert
        # so we create the vocab in the xlmr
        if self.xlm:
            self.vocab = {self.bert_tokenizer.convert_ids_to_tokens(i): i for i in range(250001)}
            self.vocab.update(self.bert_tokenizer.added_tokens_encoder)

        self.lowercase_input = False
        
    def _wordpiece_tokenize_input(self, tokens: List[str]) -> Tuple[List[str], List[int], List[int]]:
        """
        Convert a list of tokens to wordpiece tokens and offsets, as well as adding
        BERT CLS and SEP tokens to the begining and end of the sentence.

        A slight oddity with this function is that it also returns the wordpiece offsets
        corresponding to the _start_ of words as well as the end.

        We need both of these offsets (or at least, it's easiest to use both), because we need
        to convert the labels to tags using the end_offsets. However, when we are decoding a
        BIO sequence inside the SRL model itself, it's important that we use the start_offsets,
        because otherwise we might select an ill-formed BIO sequence from the BIO sequence on top of
        wordpieces (this happens in the case that a word is split into multiple word pieces,
        and then we take the last tag of the word, which might correspond to, e.g, I-V, which
        would not be allowed as it is not preceeded by a B tag).

        For example:

        `annotate` will be bert tokenized as ["anno", "##tate"].
        If this is tagged as [B-V, I-V] as it should be, we need to select the
        _first_ wordpiece label to be the label for the token, because otherwise
        we may end up with invalid tag sequences (we cannot start a new tag with an I).

        Returns
        -------
        wordpieces : List[str]
            The BERT wordpieces from the words in the sentence.
        end_offsets : List[int]
            Indices into wordpieces such that `[wordpieces[i] for i in end_offsets]`
            results in the end wordpiece of each word being chosen.
        start_offsets : List[int]
            Indices into wordpieces such that `[wordpieces[i] for i in start_offsets]`
            results in the start wordpiece of each word being chosen.
        """
        word_piece_tokens: List[str] = []
        end_offsets = []
        start_offsets = []
        cumulative = 0
        for token in tokens:
            if self.xlm:
                word_pieces = self.bert_tokenizer._tokenize(token)
            else:
                word_pieces = self.bert_tokenizer.wordpiece_tokenizer.tokenize(token)
            start_offsets.append(cumulative + 1)
            cumulative += len(word_pieces)
            end_offsets.append(cumulative)
            word_piece_tokens.extend(word_pieces)

        return word_piece_tokens, end_offsets, start_offsets

    @overrides
    def _read(self, file_path: str):
        # This reads the folds data
        
        # if `file_path` is a URL, redirect to the cache
        file_path = cached_path(file_path)
        logger.info("Reading SRL instances from dataset files at: %s", file_path)

        with codecs.open(file_path, 'r', encoding='utf8') as open_file:
            for line in open_file:
                words, tags = line.split('\t')
                words, tags = words.split(' '), tags.strip().split(' ')
                verb_indicator = [1 if label[-2:] == "-V" else 0 for label in tags]
                tokens = [Token(t) for t in words]
                yield self.text_to_instance(tokens, verb_indicator, tags)
                

            

    def text_to_instance(self,  # type: ignore
                         tokens: List[Token],
                         verb_label: List[int],
                         tags: List[str] = None) -> Instance:
        """
        We take `pre-tokenized` input here, along with a verb label.  The verb label should be a
        one-hot binary vector, the same length as the tokens, indicating the position of the verb
        to find arguments for.
        """
        # pylint: disable=arguments-differ
        metadata_dict: Dict[str, Any] = {}
        if self.bert_tokenizer is not None:
            wordpieces, offsets, start_offsets = self._wordpiece_tokenize_input([t.text for t in tokens])
            
            if self.xlm:
                wordpieces = ["<s>"] + wordpieces + ["</s>"]
            else:
                wordpieces = ["[CLS]"] + wordpieces + ["[SEP]"]
            new_verbs = _convert_verb_indices_to_wordpiece_indices(verb_label, offsets)
            metadata_dict["offsets"] = start_offsets
            # In order to override the indexing mechanism, we need to set the `text_id`
            # attribute directly. This causes the indexing to use this id.
            if self.xlm: 
                text_field = TextField([Token(t, text_id=self.vocab[t]) for t in wordpieces],
                                   token_indexers=self._token_indexers)
            else:
                text_field = TextField([Token(t, text_id=self.bert_tokenizer.vocab[t]) for t in wordpieces],
                                   token_indexers=self._token_indexers)
            
            verb_indicator = SequenceLabelField(new_verbs, text_field)

        else:
            text_field = TextField(tokens, token_indexers=self._token_indexers)
            verb_indicator = SequenceLabelField(verb_label, text_field)

        fields: Dict[str, Field] = {}
        fields['tokens'] = text_field
        fields['verb_indicator'] = verb_indicator

        if all([x == 0 for x in verb_label]):
            verb = None
            verb_index = None
        else:
            verb_index = verb_label.index(1)
            verb = tokens[verb_index].text

        metadata_dict["words"] = [x.text for x in tokens]
        metadata_dict["verb"] = verb
        metadata_dict["verb_index"] = verb_index

        if tags:
            if self.bert_tokenizer is not None:
                new_tags = _convert_tags_to_wordpiece_tags(tags, offsets)
                fields['tags'] = SequenceLabelField(new_tags, text_field)
            else:
                fields['tags'] = SequenceLabelField(tags, text_field)
            metadata_dict["gold_tags"] = tags
        fields["metadata"] = MetadataField(metadata_dict)
        return Instance(fields)


@DatasetReader.register("simple")
class SimpleReader(DatasetReader):
    """
    This data set reader is meant to read conll data, 
    preprocess it and yield it to create the folds.
    The data output is meant to be written in folds files,
    which will the be read by the SrlReader above.
    """
    def __init__(self,
                 lazy: bool = False, 
                 remove_c: bool = False) -> None:
        super().__init__(lazy)
        self.remove_c = remove_c

    @overrides
    def _read(self, file_path: str):
        # if `file_path` is a URL, redirect to the cache
        file_path = cached_path(file_path)
        PBBr_reader = PropBankBr(remove_c = self.remove_c)
        logger.info("Reading SRL instances from dataset files at: %s", file_path)

        for sentence in PBBr_reader.dataset_iterator(file_path):
            if not sentence.srl_frames:
                # Sentence contains no predicates.
                tags = ["O" for _ in sentence.words]
                words, tags = preprocess.Preprocess(sentence.words, tags).preprocess()
                yield [words, tags]
            else:
                for (_, tags) in sentence.srl_frames:
                    words, tags = preprocess.Preprocess(sentence.words, tags).preprocess()
                    yield [words, tags]



@DatasetReader.register("ud_transformer", exist_ok=True)
class UDTransformerDatasetReader(UniversalDependenciesDatasetReader):
    """
    Reads a file in the conllu Universal Dependencies format.
    This is almost identical to the original, but it transforms 
    the data in the appropriate format for a transformers model.

    # Parameters

    token_indexers : `Dict[str, TokenIndexer]`, optional (default=`{"tokens": SingleIdTokenIndexer()}`)
        The token indexers to be applied to the words TextField.
    use_language_specific_pos : `bool`, optional (default = False)
        Whether to use UD POS tags, or to use the language specific POS tags
        provided in the conllu format.
    tokenizer : `Tokenizer`, optional, default = None
        A tokenizer to use to split the text. This is useful when the tokens that you pass
        into the model need to have some particular attribute. Typically it is not necessary.
    """

    def __init__(
        self,
        token_indexers: Dict[str, TokenIndexer] = None,
        use_language_specific_pos: bool = False,
        tokenizer: Tokenizer = None,
        **kwargs,
    ) -> None:
        super().__init__(token_indexers, use_language_specific_pos, tokenizer, **kwargs)

    @overrides
    def _read(self, file_path: str):
        # if `file_path` is a URL, redirect to the cache
        file_path = cached_path(file_path)

        with open(file_path, "r", encoding="utf-8") as conllu_file:
            logger.info("Reading UD instances from conllu dataset at: %s", file_path)

            for annotation in parse_incr(conllu_file):
                # CoNLLU annotations sometimes add back in words that have been elided
                # in the original sentence; we remove these, as we're just predicting
                # dependencies for the original sentence.
                # We filter by integers here as elided words have a non-integer word id,
                # as parsed by the conllu python library.
                annotation = [x for x in annotation if isinstance(x["id"], int)]

                heads = [x["head"] for x in annotation]
                tags = [x["deprel"] for x in annotation]
                words = [x["form"] for x in annotation]
                if self.use_language_specific_pos:
                    pos_tags = [x["xpostag"] for x in annotation]
                else:
                    pos_tags = [x["upostag"] for x in annotation]
                yield self.text_to_instance(words, pos_tags, list(zip(tags, heads)))

        
    def _wordpiece_tokenize_input(self, tokens: List[str]) -> Tuple[List[str], List[int], List[int]]:
        """
        Convert a list of tokens to wordpiece tokens and offsets, as well as adding
        BERT CLS and SEP tokens to the begining and end of the sentence.

        A slight oddity with this function is that it also returns the wordpiece offsets
        corresponding to the _start_ of words as well as the end.

        We need both of these offsets (or at least, it's easiest to use both), because we need
        to convert the labels to tags using the end_offsets. However, when we are decoding a
        BIO sequence inside the SRL model itself, it's important that we use the start_offsets,
        because otherwise we might select an ill-formed BIO sequence from the BIO sequence on top of
        wordpieces (this happens in the case that a word is split into multiple word pieces,
        and then we take the last tag of the word, which might correspond to, e.g, I-V, which
        would not be allowed as it is not preceeded by a B tag).

        For example:

        `annotate` will be bert tokenized as ["anno", "##tate"].
        If this is tagged as [B-V, I-V] as it should be, we need to select the
        _first_ wordpiece label to be the label for the token, because otherwise
        we may end up with invalid tag sequences (we cannot start a new tag with an I).

        Returns
        -------
        wordpieces : List[str]
            The BERT wordpieces from the words in the sentence.
        end_offsets : List[int]
            Indices into wordpieces such that `[wordpieces[i] for i in end_offsets]`
            results in the end wordpiece of each word being chosen.
        start_offsets : List[int]
            Indices into wordpieces such that `[wordpieces[i] for i in start_offsets]`
            results in the start wordpiece of each word being chosen.
        """
        word_piece_tokens: List[str] = []
        end_offsets = []
        start_offsets = []
        cumulative = 0
        for token in tokens:
            if self.xlm:
                word_pieces = self.bert_tokenizer._tokenize(token)
            else:
                word_pieces = self.bert_tokenizer.wordpiece_tokenizer.tokenize(token)
            start_offsets.append(cumulative + 1)
            cumulative += len(word_pieces)
            end_offsets.append(cumulative)
            word_piece_tokens.extend(word_pieces)

        return word_piece_tokens, end_offsets, start_offsets


    @overrides
    def text_to_instance(
        self,  # type: ignore
        words: List[str],
        upos_tags: List[str],
        dependencies: List[Tuple[str, int]] = None,
    ) -> Instance:

        """
        # Parameters

        words : `List[str]`, required.
            The words in the sentence to be encoded.
        upos_tags : `List[str]`, required.
            The universal dependencies POS tags for each word.
        dependencies : `List[Tuple[str, int]]`, optional (default = None)
            A list of  (head tag, head index) tuples. Indices are 1 indexed,
            meaning an index of 0 corresponds to that word being the root of
            the dependency tree.

        # Returns

        An instance containing words, upos tags, dependency head tags and head
        indices as fields.
        """
        fields: Dict[str, Field] = {}

        if self.tokenizer is not None:
            wordpieces, offsets, start_offsets = self._wordpiece_tokenize_input([t.text for t in tokens])
            
            if self.xlm:
                wordpieces = ["<s>"] + wordpieces + ["</s>"]
            else:
                wordpieces = ["[CLS]"] + wordpieces + ["[SEP]"]
            if self.xlm: 
                text_field = TextField([Token(t, text_id=self.vocab[t]) for t in wordpieces],
                                   token_indexers=self._token_indexers)
            else:
                text_field = TextField([Token(t, text_id=self.bert_tokenizer.vocab[t]) for t in wordpieces],
                                   token_indexers=self._token_indexers)
        else:
            tokens = [Token(t) for t in words]
            text_field = TextField(tokens, token_indexers=self._token_indexers)
            
        fields["words"] = text_field
        fields["pos_tags"] = SequenceLabelField(upos_tags, text_field, label_namespace="pos")
        if dependencies is not None:
            # We don't want to expand the label namespace with an additional dummy token, so we'll
            # always give the 'ROOT_HEAD' token a label of 'root'.
            fields["head_tags"] = SequenceLabelField(
                [x[0] for x in dependencies], text_field, label_namespace="head_tags"
            )
            fields["head_indices"] = SequenceLabelField(
                [x[1] for x in dependencies], text_field, label_namespace="head_index_tags"
            )

        fields["metadata"] = MetadataField({"words": words, "pos": upos_tags, "offsets": start_offsets})
        return Instance(fields)




from text2story.readers import read
import re

PARENTHESES = r"\([^()]*\)"
import pandas as pd

class ReadCSV(read.Read):

    def __init__(self, columns=None):
        self.columns = columns

    def _standardize_cols(self, columns):
        std_cols_map = {}
        for c in columns:
            if len(c.strip()) > 0:
                new_col_name = re.sub(PARENTHESES,"",c)
                new_col_name = new_col_name.strip().lower()
                new_col_name = new_col_name.replace(" ","_")
                std_cols_map[c] = new_col_name
            else:
                std_cols_map[c] = ''
        return  std_cols_map
    def process(self, data_dir):
        """
        Process a set of files of a corpus

        @param string: path of data corpus

        @return: a list of tokens
        """
        pass
    def get_data(self, data_lines, **kwargs):

        for data in data_lines:
            if 'span_id' in kwargs:
                if data['span_id'] == kwargs['span_id']:
                    return data
    def process_file(self, data_file):
        """
        Process only one file of a corpus.
        """
        df = pd.read_csv(data_file)
        col_names_nonstd = df.columns.tolist()
        col_names_map = self._standardize_cols(col_names_nonstd)

        data_lst = []

        for idx, row in df.iterrows():
            data = {}
            for col_name in  col_names_map:

                if self.columns:
                    if col_names_map[col_name] in self.columns:
                        data[col_names_map[col_name]] = row[col_name]
                else:
                    if col_name != '':
                        data[col_names_map[col_name]] = row[col_name]
            if data != {}:
                data_lst.append(data)


        return data_lst



import re
import sys

from text2story.readers.token_corpus import TokenCorpus
from text2story.core.utils import join_tokens

class NodeEvent:

    def __init__(self, event_txt, event_class, event_idx, event_arg, event_sent, id_ann = None):
        self.txt = event_txt
        self.class_type = event_class
        self.idx = event_idx 
        self.argn = event_arg
        self.sent_id = event_sent
        self.id_ann = id_ann

class EdgeEvent:

	def __init__(self, edge_type):
		self.edge_type = edge_type
		self.nodes = []

	def add_node(self, event_txt, event_class, event_idx, event_arg, event_sent, id_ann = None):
		self.nodes.append(NodeEvent(event_txt, event_class, event_idx, event_arg, event_sent, id_ann))

def sieve_bbubbles_events(event_lst, type_big_bubble):
    """
    A sieve for events that has temporal links of identity types
    """

    ans = {}
    not_ans = {}

    for id_ann in event_lst:
        
        for tok in event_lst[id_ann]:


            for attr_item in tok.attr:
 
                ann_type = attr_item[0]
                attr_map = attr_item[1]

                if "Class" not in attr_map:
                    continue

                if attr_map["Class"] == type_big_bubble:
                    is_big_bubble = False

                    for r in tok.relations:
                        #print(r.rel_type, id_ann, tok.text)
                        if is_event(r.toks[0]):
                            if r.rel_type == "TLINK_identity" and\
                                id_ann not in ans and type_big_bubble in get_event_type(r.toks[0]):

                                #toks_ids = [rel_toks.id_ann for rel_toks in r.toks]
                                #print("id_ann: ",id_ann, toks_ids)
                                ans[id_ann] = event_lst[id_ann]
                                is_big_bubble = True
                                break

                    if not(is_big_bubble):
                        if id_ann in not_ans:
                            not_ans[id_ann].append(tok)
                        else: 
                            not_ans[id_ann] = [tok]
                    
                else:
                    if id_ann in not_ans:
                        not_ans[id_ann].append(tok)
                    else: 
                        not_ans[id_ann] = [tok]

    return ans, not_ans


def get_reporting_events(tok_lst):

    event_lst = {}


    for i in range(len(tok_lst)):
        tok = tok_lst[i]


        for attr_item in tok.attr:

            ann_type = attr_item[0]
            attr_map = attr_item[1]

            if "Class" in attr_map and attr_map["Class"] == "Reporting": # it is an event token

                for id_ann in tok.id_ann:

                    if id_ann in event_lst:
                        event_lst[id_ann].append(tok)
                    else:
                        event_lst[id_ann] = [tok]
                break
    return event_lst

def get_offset_lst(tok_lst):
    res = []
    for tok in tok_lst:
        res.append((tok.offset, tok.offset + len(tok.text), tok.id_ann))
    return res
def get_first_event(tok_lst):
    offset_lst = get_offset_lst(tok_lst)
    offset_lst.sort(key=lambda x:x[0])

    return offset_lst[0]
def get_last_event(tok_lst):
    offset_lst = get_offset_lst(tok_lst)
    offset_lst.sort(key=lambda x: x[0])
    return offset_lst[-1]

def get_nested_events_nonlinked(event_lst, nested_events_ids):
    """
    Get nested events in reporting events that are not directly linked to it
    @param event_lst: the list of events that are nested in a given reporting event
    @param nested_events_ids: the ids of events that are directly connected to reporting events
    @return:
    """
    # For each event E in event_lst, check if there is an Et that
    # it is the same sentence that E, and is linked with TLINK_identity
    # Also, Et should not be a reporting event
    X =[]
    for E in event_lst:
        for rel in E.relations:
            if rel.rel_type == "TLINK_identity":
                Et = rel.toks[0]
                if Et.sent_id == E.sent_id and\
                        Et.get_attr_value('Class') != 'Reporting' and\
                        Et.id_ann[0] not in nested_events_ids:
                    X.append(Et)
                    nested_events_ids.add(Et.id_ann[0])
    return X

def get_embedded_events(doc):

    event_lst = get_all_events_sent(doc)

    embedded_events_lst = {}
    embedded_events_ids = set()
    count_supressed_reporting_events = 0

    for sent_id in event_lst:

        reporting_events = []
        embedded_events = []

        for event in event_lst[sent_id]:
            if event.get_attr_value("Class") == "Reporting":
                reporting_events.append(event)
            else:
                embedded_events.append(event)

        if len(reporting_events) > 1:
            # in this situation there is a chain of reporting events
            # some of them are concerning of story, and others are
            # related to
            print(f"{len(reporting_events)} reporting events in {sent_id}")
            count_supressed_reporting_events += len(reporting_events) - 1

            # TODO: for now just ignoring other reporting events and considering only one
            # considering the ROOT
            for e in embedded_events:
                embedded_events_ids.add(e.id_ann[0])
            embedded_events_lst[sent_id] = (reporting_events[0], embedded_events)


        elif len(reporting_events) == 1:
            for e in embedded_events:
                embedded_events_ids.add(e.id_ann[0])
            embedded_events_lst[sent_id] = (reporting_events.pop(), embedded_events)

    return embedded_events_lst, embedded_events_ids, count_supressed_reporting_events


def get_nested_events_reporting(doc):
    # for each id annotation get the related events
    event_lst = get_all_events(doc)
    nested_events = {}
    nested_events_ids = set()

    for event_id in event_lst:

        for event in event_lst[event_id]:
            event_class = event.get_attr_value('Class', 'Event')

            if event_class == 'Reporting':
                sent_id = event.sent_id

                # devo olhar somente para quem ele se relaciona? E se eu fizer a coleta por sentença?
                # fazer um get_all_events por sentença?
                for r in event.relations:
                    tok_rel = r.toks[0]
                    # print("-->", tok_rel.id_ann)
                    if tok_rel.is_type("Event"):
                        if tok_rel.get_attr_value('Class') == 'Reporting' and r.rel_type == 'TLINK_identity':
                            continue

                        rel_sent_id = tok_rel.sent_id
                        if rel_sent_id == sent_id and tok_rel.id_ann[0] not in nested_events_ids:
                            if event_id in nested_events:
                                nested_events[event_id].append(tok_rel)
                            else:
                                nested_events[event_id] = [tok_rel]
                            nested_events_ids.add(tok_rel.id_ann[0])

    # some events are nested in a reporting event, but are not directly linked to it
    for event_id in nested_events:
        non_linked = get_nested_events_nonlinked(nested_events[event_id], nested_events_ids)
        nested_events[event_id] += non_linked

    return event_lst, nested_events, nested_events_ids


def get_all_events_sent(tok_lst):
    """
    Get all events from a document and index them by the sentence number
    @param tok_lst:
    @return:
    """

    event_lst = {}
    event_id_lst = set()

    for i in range(len(tok_lst)):
        tok = tok_lst[i]

        if tok.is_type("Event") and tok.id_ann[0] not in event_id_lst:
            if tok.sent_id in event_lst:
                event_lst[tok.sent_id].append(tok)
            else:
                event_lst[tok.sent_id] = [tok]
            event_id_lst.add(tok.id_ann[0])

    return event_lst

def get_all_events(tok_lst):

    event_lst = {}


    for i in range(len(tok_lst)):
        tok = tok_lst[i]


        for attr_item in tok.attr:

            ann_type = attr_item[0]
            attr_map = attr_item[1]

            if "event" in ann_type.lower():

                for id_ann in tok.id_ann:

                    if id_ann in event_lst:
                        event_lst[id_ann].append(tok)
                    else:
                        event_lst[id_ann] = [tok]
                break
            	

    return event_lst

def get_txt_events(event_lst):

    event_txt_lst = []

    for id_ann in event_lst:
        text_tokens = [e.text for e in event_lst[id_ann]]
        event_txt = " ".join(text_tokens)
        event_txt_lst.append(event_txt)

    return event_txt_lst

        
def get_hist_report_events(tok_lst):

    event_lst = get_all_events(tok_lst)
    hist_reporting = {}
    for e in event_lst:

        rel_type = get_event_type(e)
        if "Reporting" in rel_type:
            tok_event = e.text.lower()
            if tok_event in hist_reporting:
                hist_reporting[tok_event] += 1
            else:
                hist_reporting[tok_event] = 1
    return hist_reporting

def map_events_bysent(tok_lst):

    event_lst = get_all_events(tok_lst)
    map_sent = {}
    for e in event_lst:

        primary_type = get_event_type(e)[0]

        if e.sent_id in map_sent:
            if primary_type in map_sent[e.sent_id]:
                map_sent[e.sent_id][primary_type] += 1
            else:
                map_sent[e.sent_id][primary_type] = 1
        else:
            map_sent[e.sent_id] = {primary_type:1}

    return map_sent

def is_event(event):

    for attr_item in event.attr:

        ann_type = attr_item[0]
        attr_map = attr_item[1]

        if ann_type.lower() == "event":
            return True
    return False


def get_event_type(event):

    event_type_lst = []

    
    for attr_item in event.attr:

        ann_type = attr_item[0]
        attr_map = attr_item[1]

        if "Class" in attr_map: 
            event_type_lst.append(attr_map["Class"])
                
    #print("-->", event_type_lst)
    return event_type_lst



"""
	text2story.core.entity_structures

	Entity structures classes (participant, TimeX and Event)
"""
import sys


class ParticipantEntity:
    """
    Representation of an Participant entity.

    Attributes
    ----------
    text: str
        The textual representation of the participant.
    character_span: tuple[int, int]
        The character span of the participant.
    lexical_head: str
        The lexical head of the participant.
        Possible values are: 'Noun' or 'Pronoun'.
    type: str
        The type of the participant.
        Possible values are: 'Per', 'Org', 'Loc', 'Obj', 'Nat' or 'Other'.
    individuation: str
        Stipulation of whether the participant is a set, a single individual, or a mass quantity.
        Possible values are: 'Set', 'Individual' or 'Mass'.
        NOTE: For now, using the label 'Individual' to all participants.
    involvement: str
        The specification of how many entities or how much of the domain are/is participating in an event.
        Possible values are: '0', '1', '>1', 'All' or 'Und'.
        NOTE: For now, using the label '1' to all participants.
    """

    def __init__(self, text, character_span, lexical_head, participant_type):
        self.text = text
        self.character_span = character_span
        self.lexical_head = lexical_head
        self.type = participant_type
        self.individuation = 'Individual'
        self.involvement = '1'


class TimeEntity:
    """
    Representation of a time entity.

    Attributes
    ----------
    text: str
        The textual representation of the time.
    character_span: tuple[int, int]
        The character span of the time.
    value: str
        The value of the time.
        Possible values are: value ::=  Duration | Date | Time | WeekDate | WeekTime | Season | PartOfYear | PaPrFu(Past Present Future Reference)
    type: str
        The type of the time.
        Possible values are: 'Date', 'Time', 'Duration' and 'Set'.
    temporal_function: str
        Possible values are: 'None' or 'Publication_Time'.
    """

    def __init__(self, text, character_span, timex_type, temporal_function='Publication_Time'):
        self.text = text
        self.character_span = character_span
        #self.value = value
        self.type = timex_type
        self.temporal_function = temporal_function


class EventEntity:
    """
    Representation of an event entity.
        TODO: Annotations (A)
    """

    def __init__(self, text, character_span,**kwargs):
        self.text = text
        self.character_span = character_span

        for attr_name, attr_value in kwargs.items():
            if attr_name.lower() == "event_class":
                setattr(self, "Class", attr_value)
            elif attr_name.lower() == "event_type":
                setattr(self, "Event_Type", attr_value)
            else:
                setattr(self,attr_name, attr_value)

        if hasattr(self, "Event_Type"):
            event_type = self.__getattribute__("Event_Type")
            # deafult value for event_type
            setattr(self, "Event_Type","Process")

        #self.factuality = "Factual"
        #self.tense = "Pres"

class SpatialRelationEntity:
    def __init__(self, text, character_span,**kwargs):
        self.text = text
        self.character_span = character_span

        for attr_name, attr_value in kwargs.items():
            setattr(self,attr_name, attr_value)




from transformers import PretrainedConfig

class SRLModelConfig(PretrainedConfig):
    model_type = "srl"
    
    def __init__(
        self,
        num_labels=0,
        bert_model_name="bert-base-uncased",
        embedding_dropout=0.0,
        label2id = {},
        id2label = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_labels = num_labels
        self.bert_model_name = bert_model_name
        self.embedding_dropout = embedding_dropout
        self.label2id = label2id
        self.id2label = id2label

    def to_dict(self):
        config_dict = super().to_dict()

        config_dict["num_labels"] = self.num_labels
        # config_dict["bert_model_name"] = self.bert_model_name
        # config_dict["embedding_dropout"] = self.embedding_dropout

        return config_dict



import json
import random
import re
from pathlib import Path

# import nltk
from nltk.tokenize import sent_tokenize
# nltk.download('punkt')
from text2story.brat2viz.brat2drs.brat2drs import get_files, read_file
from text2story.brat2viz.brat import file_parser


def get_all_entities(ann_dict):
    entities_dict = {}

    # Avoids >1 dict entries of the same entity (O Comando Metropolitano vs O Comando Metropolitano do Porto)
    tag_ref_beginning = {}

    for x in ann_dict:
        if x['ann_type'] == 'TextBound' and 'attribute' in x:
            if x.get('tag_ref') is None and x.get('s_pos') is not None:
                x['tag_ref'] = x['s_pos']
            try:
                if x['tag_ref'] not in tag_ref_beginning.keys():
                    tag_ref_beginning[x['tag_ref']] = x['tag_id']
                    entities_dict[x['tag_id']] = {'attribute': x['attribute'], 'value': x['value'],
                                              'tag_ref_beginning': x['tag_ref']}
            except TypeError as e:
                print(x, tag_ref_beginning)
                raise e
        # Completes the information in the dict
        elif x['ann_type'] == 'Attribute':
            if x['tag_ref'] in entities_dict.keys():
                entities_dict[x['tag_ref']].update({x['attr_type']: x['value']})


        # Finds Objectal links
        # Considers "OLINK_objIdentity" synonyms
        # Considers "OLINK_partOf", "OLINK_subset" and "OLINK_memberOf" as parts of a whole
        # Ignores "OLINK_referentialDisjunction"
        # Finds synonyms
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_objIdentity':
            if entities_dict.get(x['tag_ref1']) is not None:
                if entities_dict[x['tag_ref1']].get('synonyms') is not None:
                    entities_dict[x['tag_ref1']]['synonyms'] += ';' + x['tag_ref2']
                else:
                    entities_dict[x['tag_ref1']]['synonyms'] = x['tag_ref2']

        # Finds "part of" relations
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_partOf':
            if entities_dict.get(x['tag_ref1']) is not None:
                if entities_dict[x['tag_ref1']].get('Part_Of') is not None:
                    entities_dict[x['tag_ref1']]['Part_Of'] += ';' + x['tag_ref2']
                else:
                    entities_dict[x['tag_ref1']]['Part_Of'] = x['tag_ref2']

        # Finds "member of" relations
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_memberOf':
            if entities_dict.get(x['tag_ref1']) is not None:
                if entities_dict[x['tag_ref1']].get('Part_Of') is not None:
                    entities_dict[x['tag_ref1']]['Part_Of'] += ';' + x['tag_ref2']
                else:
                    entities_dict[x['tag_ref1']]['Part_Of'] = x['tag_ref2']

        # Finds "subset" relations
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_subset':
            if entities_dict.get(x['tag_ref1']) is not None:
                if entities_dict[x['tag_ref1']].get('Part_Of') is not None:
                    entities_dict[x['tag_ref1']]['Part_Of'] += ';' + x['tag_ref2']
                else:
                    entities_dict[x['tag_ref1']]['Part_Of'] = x['tag_ref2']

    return entities_dict


def get_participants(ann_dict):
    participants_dict = {}

    # Avoids >1 dict entries of the same entity (O Comando Metropolitano vs O Comando Metropolitano do Porto)
    tag_ref_beginning = {}
    status = ''

    for x in ann_dict:
        if x.get('attribute') is not None:
            status = x.get('attribute')

        # Creates a dict of participants for fast lookup
        if x['ann_type'] == 'TextBound' and 'attribute' in x \
                and (x['attribute'] == 'ACTOR' or x['attribute'] == 'Participant'):
            status = 'Participant'
            if x.get('tag_ref') is None and x.get('s_pos') is not None:
                x['tag_ref'] = x['s_pos']
            if x['tag_ref'] not in tag_ref_beginning.keys():
                tag_ref_beginning[x['tag_ref']] = x['tag_id']
                participants_dict[x['tag_id']] = {'value': x['value'], 'tag_ref_beginning': x['tag_ref']}
            #
            # else:  # assume synonym if the entity starts in the same character as another
            # or just ignore it
            #    if participants_dict[tag_ref_beginning[x['tag_ref']]].get('synonyms') is not None:
            #        participants_dict[tag_ref_beginning[x['tag_ref']]]['synonyms'] += ', ' + x['tag_id']
            #    else:
            #        participants_dict[tag_ref_beginning[x['tag_ref']]]['synonyms'] = x['tag_id']

        # Completes the information of the participants in the dict
        elif x['ann_type'] == 'Attribute' and status == "Participant":
            if x['tag_ref'] in participants_dict.keys():
                participants_dict[x['tag_ref']].update({x['attr_type']: x['value']})

        # Finds Objectal links
        # Considers "OLINK_objIdentity" synonyms
        # Considers "OLINK_partOf", "OLINK_subset" and "OLINK_memberOf" as parts of a whole
        # Ignores "OLINK_referentialDisjunction"
        # Finds synonyms
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_objIdentity':
            if participants_dict.get(x['tag_ref1']) is not None:
                if participants_dict[x['tag_ref1']].get('synonyms') is not None:
                    participants_dict[x['tag_ref1']]['synonyms'] += ', ' + x['tag_ref2']
                else:
                    participants_dict[x['tag_ref1']]['synonyms'] = x['tag_ref2']

        # Finds "part of" relations
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_partOf':
            if participants_dict.get(x['tag_ref1']) is not None:
                if participants_dict[x['tag_ref1']].get('Part_Of') is not None:
                    participants_dict[x['tag_ref1']]['Part_Of'] += ', ' + x['tag_ref2']
                else:
                    participants_dict[x['tag_ref1']]['Part_Of'] = x['tag_ref2']

        # Finds "member of" relations
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_memberOf':
            if participants_dict.get(x['tag_ref1']) is not None:
                if participants_dict[x['tag_ref1']].get('Part_Of') is not None:
                    participants_dict[x['tag_ref1']]['Part_Of'] += ', ' + x['tag_ref2']
                else:
                    participants_dict[x['tag_ref1']]['Part_Of'] = x['tag_ref2']

        # Finds "subset" relations
        elif x['ann_type'] == 'Relation' and x['rel_type'] == 'OLINK_subset':
            if participants_dict.get(x['tag_ref1']) is not None:
                if participants_dict[x['tag_ref1']].get('Part_Of') is not None:
                    participants_dict[x['tag_ref1']]['Part_Of'] += ', ' + x['tag_ref2']
                else:
                    participants_dict[x['tag_ref1']]['Part_Of'] = x['tag_ref2']

    return participants_dict


def get_events(f_parser):
    events_dict = {}
    status = ''
    for x in f_parser:
        if x.get('attribute') is not None:
            status = x.get('attribute')

        # Creates a dict of events for fast lookup
        if x['ann_type'] == 'TextBound' and 'attribute' in x and x['attribute'] == 'Event':
            status = 'Event'
            events_dict[x['tag_id']] = {'value': x['value']}

        # Completes the information of the events in the dict
        elif x['ann_type'] == 'Attribute' and status == "Event":
            if x['tag_ref'] in events_dict.keys():
                events_dict[x['tag_ref']].update({x['attr_type']: x['value']})
    return events_dict


def get_semantic_relations(f_parser):
    sr_dict = {}
    for item in f_parser:
        if item['ann_type'] == 'Relation' and item['rel_type'][0:6] == 'SRLINK':
            sr_dict[item['tag_id']] = {'rel_type': item['rel_type'], 'tag_ref1': item['tag_ref1'],
                                       'tag_ref2': item['tag_ref2']}
    return sr_dict


def get_all_relations(f_parser):
    link_dict = {}
    for item in f_parser:
        if item['ann_type'] == 'Relation':
            rel_type = item['rel_type'].split('_')[0]
            if 'LINK' in rel_type:
                link_dict[item['tag_id']] = {'rel_type': item['rel_type'], 'tag_ref1': item['tag_ref1'],
                                             'tag_ref2': item['tag_ref2']}
    return link_dict


def get_all_times(f_parser):
    time_dict = {}
    status = ''
    for x in f_parser:
        if status == 'Time':
            if x.get('attribute') != None:
                status = x.get('attribute')

        # Creates a dict of times for fast lookup
        if x['ann_type'] == 'TextBound' and 'attribute' in x and x['attribute'] == 'Time':
            time_dict[x['tag_id']] = {'time': x['value'], 'attribute': x['attribute'], 'tag_ref': x['tag_ref']}
            status = 'Time'

        # Completes the information of the times in the dict
        elif x['ann_type'] == 'Attribute' and status == 'Time':
            if x['tag_ref'] in time_dict.keys():
                time_dict[x['tag_ref']].update({x['attr_type']: x['value']})

    return time_dict


def filter_dict(entities, attribute, value_list):
    return {key: item for (key, item) in entities.items() if item.get(attribute) in value_list}


def group_qslinks_relations(qslink_relations, spatial_relations):
    grouped_relations = {}
    for (k, v) in spatial_relations.items():
        if v.get('Topological') is not None:
            grouped_relations[k] = {'value': v['value'], 'tag_ref_beginning': int(v['tag_ref_beginning']),
                                    'Topological': v['Topological']}
    for (k, v) in qslink_relations.items():
        if v['tag_ref1'] in grouped_relations.keys():
            grouped_relations[v['tag_ref1']].update({v['rel_type']: v['tag_ref2']})
    return grouped_relations


def get_event_time_tlinks(tlinks_relations, events, all_times):
    results = {}
    for tlink_key, tlink_value in tlinks_relations.items():
        if tlink_value['tag_ref1'] in events.keys() and tlink_value['tag_ref2'] in all_times.keys():
            # just for viz
            tlink_value.update({'event': tlink_value['tag_ref1']})
            tlink_value.update({'time': tlink_value['tag_ref2']})
            tlink_value.update({tlink_value['tag_ref1']: events[tlink_value['tag_ref1']]})
            tlink_value.update({tlink_value['tag_ref2']: all_times[tlink_value['tag_ref2']]})
            del tlink_value['tag_ref1']
            del tlink_value['tag_ref2']

            results[tlink_key] = tlink_value

        elif tlink_value['tag_ref1'] in all_times.keys() and tlink_value['tag_ref2'] in events.keys():
            # just for viz
            tlink_value.update({'time': tlink_value['tag_ref1']})
            tlink_value.update({'event': tlink_value['tag_ref2']})
            tlink_value.update({tlink_value['tag_ref1']: all_times[tlink_value['tag_ref1']]})
            tlink_value.update({tlink_value['tag_ref2']: events[tlink_value['tag_ref2']]})
            del tlink_value['tag_ref1']
            del tlink_value['tag_ref2']

            results[tlink_key] = tlink_value
    return results


def group_entities(characters_dict, entities_dict):
    participants_to_del = set()

    for key in characters_dict.keys():
        # Groups part/whole relations
        wholes_joined = characters_dict[key].get("Part_Of")
        if wholes_joined is not None:
            wholes_list = wholes_joined.split(';')
            wholes_list = [item.strip() for item in wholes_list]

            for whole in wholes_list:
                if characters_dict.get(whole) is not None:
                    # print(whole)
                    if characters_dict[whole].get("Parts") is not None:
                        check_unique_list = characters_dict[whole]["Parts"].split(";")
                        check_unique_list = [item.strip() for item in check_unique_list]
                        if key not in check_unique_list:
                            characters_dict[whole]["Parts"] += ";" + key
                    else:
                        characters_dict[whole]["Parts"] = key

            # Falta fazer todas as partes recursivamente
            participants_to_del.add(key)

    for key in reversed(characters_dict.keys()):
        # Groups all synonyms
        synonyms = characters_dict[key].get("synonyms")
        if synonyms is not None:
            synonyms_list = synonyms.split(';')
            synonyms_list = [item.strip() for item in synonyms_list]

            # Gets synonyms of synonyms
            syn_queue = []
            for syn in synonyms_list:
                syn_queue.append(syn)
            while syn_queue:
                if syn_queue[0] != key and syn_queue[0] not in participants_to_del:
                    if characters_dict.get(syn_queue[0]):
                        new_syns = entities_dict[syn_queue[0]].get("synonyms")
                    elif entities_dict.get(syn_queue[0]):
                        new_syns = entities_dict[syn_queue[0]].get("synonyms")
                    if new_syns is not None:
                        new_syns_list = new_syns.split(';')
                        for new_syn_item in new_syns_list:
                            if new_syn_item not in synonyms_list:
                                syn_queue.append(new_syn_item)
                                synonyms_list.append(new_syn_item)
                syn_queue.pop(0)

            for syn_item in synonyms_list:
                fields_to_check = ["Individuation_Domain", "Involvement", "Lexical_Head", "Participant_Type_Domain",
                                   "Parts", "value"]
                # Add the synonym field to the original character dict, if different from original
                # If field == value, add if not present
                for field in fields_to_check:
                    if characters_dict.get(syn_item) is not None:
                        if characters_dict[syn_item].get(field) is not None:
                            check_unique_list = characters_dict[syn_item][field].split(";")
                            check_unique_list = [item.strip() for item in check_unique_list]
                            if characters_dict[key].get(field) is not None and \
                                    characters_dict[key][field] not in check_unique_list:
                                characters_dict[syn_item][field] += ";" + characters_dict[key][field]
                        elif field == "Parts":
                            if characters_dict[key].get(field) is not None:
                                characters_dict[syn_item][field] = characters_dict[key][field]

            participants_to_del.add(key)

    for key in characters_dict.copy().keys():
        if key in participants_to_del:
            del characters_dict[key]

    return characters_dict


def write_jsons(chars_dict, all_characters, file_path, txt_files_dir, output_dir, titles_list, events, locations,
                spatial_info, all_times, event_time_tlinks):
    final_json = {}
    final_json_extended = {}
    title = re.split('\\.|\\/|\\\\', file_path)[-2]

    # Organize character section
    characters = []
    list_of_locations = []
    list_of_dates = []
    for key, value in chars_dict.items():
        rgb = random.sample(range(256), 3)
        rgb_str = 'rgb(' + str(rgb[0]) + ',' + str(rgb[1]) + ',' + str(rgb[2]) + ')'

        name = value["value"].split(';')[0].strip()
        new_character = {"id": key, "name": name, "affiliation": rgb_str}

        if len(value["value"].split(';')) > 1:
            new_character["synonyms"] = [item.strip() for item in value["value"].split(';')[1:] if item != name]
            new_character["synonyms"] = list(dict.fromkeys(new_character["synonyms"]))
        if value.get("Parts") is not None:
            if len(value["Parts"].split(';')) > 0:
                new_character_parts = [item.strip() for item in value["Parts"].split(';')]
                for part in new_character_parts:
                    if new_character.get("synonyms") is not None:
                        new_character["synonyms"] += [item.strip() for item in all_characters[part]["value"].split(';')
                                                      if item != name]
                    else:
                        new_character["synonyms"] = [item.strip() for item in all_characters[part]["value"].split(';')
                                                     if item != name]
        characters.append(new_character)

    for (k, v) in locations.items():
        v['id'] = k
        v['value'] = v['value'].split(';')[0].strip()

    for (k, v) in all_times.items():
        v['id'] = k

    # Organize scene section
    txt_file = txt_files_dir + "/" + title + ".txt"
    with open(txt_file, "r", encoding="utf8") as f:
        unprocessed_txt = f.read()

    sentences_list = sent_tokenize(unprocessed_txt)

    # Clean first sentence (remove headings)
    first_sentence = sentences_list[0].split('\n')[3:]
    char_accumulator = len('\n'.join(sentences_list[0].split('\n')[0:3])) + 1
    sentences_list[0] = '\n'.join(first_sentence)

    sentences_dict = {}
    for i, sentence in enumerate(sentences_list):
        sentences_dict[i] = {'value': sentence, 'beginning_char_index': char_accumulator,
                             'end_char_index': char_accumulator + len(sentence)}
        char_accumulator += len(sentence)

    ## Normal file
    scenes = []
    for k, sentence in sentences_dict.items():
        character_list = []
        for c in characters:
            if c["name"] in sentence['value'] and c["id"] not in character_list:
                character_list.append(c["id"])
            if c.get("synonyms"):
                for synonym in c['synonyms']:
                    if synonym in sentence['value'] and c["id"] not in character_list:
                        character_list.append(c["id"])
        location_list = []
        for loc_key, loc_value in locations.items():
            if loc_value['value'] in sentence['value'] and \
                    sentence['beginning_char_index'] < int(loc_value['tag_ref_beginning']) <= sentence['end_char_index'] \
                    and loc_key not in location_list:
                location_list.append(loc_key)
        times_list = []
        for time_key, time_value in all_times.items():
            if time_value['time'] in sentence['value'] and \
                    sentence['beginning_char_index'] < int(time_value['tag_ref']) <= sentence['end_char_index'] and \
                    time_key not in times_list:
                times_list.append(time_key)

        new_scene = {"characters": character_list, "description": sentence['value'],
                     "title": sentence['value'], "date": times_list,
                     "location": location_list}
        scenes.append(new_scene)

    ## Extended file
    ordered_titles_list = {}
    scene_events = []
    for scene_key, scene_title in titles_list.items():
        events_keys_in_scene = scene_title['events'].split(';')
        event_position_index = events[events_keys_in_scene[0]]['tag_ref_beginning']
        scene_events.append((int(event_position_index), scene_key))
    scene_events_sorted = sorted(scene_events)
    for (pos, key) in scene_events_sorted:
        ordered_titles_list[key] = titles_list[key]

    extended_scenes = []
    for scene_title in ordered_titles_list.values():
        finished = False
        description = None
        title_words = scene_title['title'].split(' ')
        for k, sentence in sentences_dict.items():
            if all(word in sentence['value'] for word in title_words):
                finished = True
                description = sentence['value']
            if finished is True:
                break
        if finished is True and description is not None:
            character_list = []
            for c in characters:
                if c["name"] in scene_title['title'] and c["id"] not in character_list:
                    character_list.append(c["id"])
                if c.get("synonyms"):
                    for synonym in c['synonyms']:
                        if synonym in scene_title['title'] and c["id"] not in character_list:
                            character_list.append(c["id"])
            location_list = []
            for qs in spatial_info.values():
                if (qs.get('QSLINK_figure') in scene_title['events'].split(';') or
                    qs.get('QSLINK_figure') in scene_title['participants'].split(';') or
                    qs.get('QSLINK_figure') in location_list) and \
                        qs.get('QSLINK_ground') in locations.keys():
                    location_list.append(qs['QSLINK_ground'])
            times_list = []
            for time_key, time_value in event_time_tlinks.items():
                if time_value['event'] in scene_title['events'].split(',') and time_value['time'] not in times_list:
                    times_list.append(time_value['time'])

            new_scene = {"characters": character_list, "description": description, "title": scene_title['title'],
                         "location": location_list, "date": times_list}
            extended_scenes.append(new_scene)

    for (k, v) in locations.items():
        v.pop('tag_ref_beginning', None)
        v.pop('Lexical_Head', None)
        v.pop('Individuation_Domain', None)
        v.pop('Participant_Type_Domain', None)
        v.pop('Involvement', None)
        v.pop('attribute', None)
        list_of_locations.append(v)

    time_attributes = ['time', 'value', 'id', 'TemporalFunction']
    for (k, v) in all_times.items():
        attr_to_del = []
        for (kv, vv) in v.items():
            if kv not in time_attributes:
                attr_to_del.append(kv)
        for ka in attr_to_del:
            del v[ka]
        list_of_dates.append(v)

    # Build final dict
    final_json['title'] = title
    final_json['characters'] = characters
    final_json['locations'] = list_of_locations
    final_json['dates'] = list_of_dates
    final_json['scenes'] = scenes

    # Build final extended dict
    final_json_extended['title'] = title
    final_json_extended['characters'] = characters
    final_json_extended['locations'] = list_of_locations
    final_json_extended['dates'] = list_of_dates
    final_json_extended['scenes'] = extended_scenes

    # Dump dict into file
    with open(output_dir + title + ".json", "w") as outfile:
        outfile.write(json.dumps(final_json, indent=4))

    # Dump extended dict into file
    with open(output_dir + title + "_extended.json", "w") as outfile_extended:
        outfile_extended.write(json.dumps(final_json_extended, indent=4))


# [lusa] Considering characters all participants of:
#   Individuation_Domain = Individual, Set
#   Participant_Type_Domain = Pl_*
#   Lexical_Head = Noun
# def get_locations(participants):
#     type_domains = ['Pl_water', 'Pl_celestial', 'pl_mountain', 'Pl_civil', 'Pl_country', 'Pl_mount_range',
#                     'Pl_capital', 'Pl_region', 'Pl_state']
#     return {key: value for (key, value) in participants.items() if
#             (value.get('Individuation_Domain') == 'Individual' or value.get('Individuation_Domain') == 'Set')
#             and (value.get('Participant_Type_Domain') in type_domains) and value.get('Lexical_Head') == 'Noun'}


def filter_semantic_relations(relations_to_consider, all_relations):
    return {key: value for (key, value) in all_relations.items() if value['rel_type'][-len(value['rel_type']) + 7:]
            in relations_to_consider}


def link_sr_to_info(considered_sr_relations, participants, events):
    sr_relations_info = considered_sr_relations.copy()
    for (k, v) in sr_relations_info.items():
        if v['tag_ref1'] in participants:
            sr_relations_info[k].update({'tag_ref1_value': participants[v['tag_ref1']]['value']})
        elif v['tag_ref1'] in events:
            sr_relations_info[k].update({'tag_ref1_value': events[v['tag_ref1']]['value']})

        if v['tag_ref2'] in participants:
            sr_relations_info[k].update({'tag_ref2_value': participants[v['tag_ref2']]['value']})
        elif v['tag_ref2'] in events:
            sr_relations_info[k].update({'tag_ref2_value': events[v['tag_ref2']]['value']})
    return sr_relations_info


def assemble_extended_scenes(semantic_roles_dict_with_dups, participants, events):
    results = {}

    # Deleting duplicates
    semantic_roles_dict = {}
    for sr_key, sr_value in semantic_roles_dict_with_dups.items():
        if sr_value not in semantic_roles_dict.values():
            semantic_roles_dict[sr_key] = sr_value

    if len(semantic_roles_dict) == 0:
        return {}

    semantic_roles = iter(semantic_roles_dict.items())
    k, v = next(semantic_roles)
    prev_k = None
    prev_v = None
    current_scene = {}

    # primeiro elemento a ser feito cá fora
    if v['tag_ref1'] in events:  # verbo -> escrito em segundo
        if v['tag_ref2'] in participants or (v['tag_ref2'] in events and events[v['tag_ref2']]['Pos'] == 'Noun'):
            current_scene['relations'] = {k: v}
            current_scene['events'] = v['tag_ref1']
            current_scene['participants'] = v['tag_ref2']
    elif v['tag_ref1'] in participants or (v['tag_ref1'] in events and events[v['tag_ref1']]['Pos'] == 'Noun'):
        if v['tag_ref2'] in events:
            current_scene['relations'] = {k: v}
            current_scene['events'] = v['tag_ref2']
            current_scene['participants'] = v['tag_ref1']

    beginning_breakpoints = ['SRLINK_pivot', 'SRLINK_agent', 'SRLINK_patient']
    end_breakpoints = ['SRLINK_result']
    double_check = {}
    finished = False
    while not finished:
        (next_k, next_v) = next(semantic_roles)
        relation_processed = False

        if v['rel_type'] in beginning_breakpoints:
            if prev_v is not None and current_scene != {} and \
                    ((prev_v['rel_type'] != 'SRLINK_agent' or v['rel_type'] != 'SRLINK_patient') and
                     (prev_v['rel_type'] != 'SRLINK_patient' or v['rel_type'] != 'SRLINK_agent')):

                # Check for matches in the previous scenes
                is_event = False
                is_participant = False
                scene_id = ''
                for kres, res in results.items():
                    if v['tag_ref1'] == list(res['relations'].values())[-1]['tag_ref1']:  # normally events
                        is_event = True
                        scene_id = kres
                    elif v['tag_ref1'] == list(res['relations'].values())[-1]['tag_ref2']:  # normally nouns
                        is_participant = True
                        scene_id = kres
                if is_event and scene_id != '':
                    if v['tag_ref2'] in participants or (
                            v['tag_ref2'] in events and
                            events[v['tag_ref2']].get('Pos') == 'Noun'):
                        results[scene_id]['relations'].update({k: v})
                        results[scene_id]['participants'] += ';' + v['tag_ref2']
                        relation_processed = True
                elif is_participant and scene_id != '':
                    if v['tag_ref2'] in participants or (
                            v['tag_ref2'] in events and
                            events[v['tag_ref2']].get('Pos') == 'Noun'):
                        results[scene_id]['relations'].update({k: v})
                        results[scene_id]['participants'] += ';' + v['tag_ref2']
                        relation_processed = True

                # Finish current scene and start the next one
                current_scene['key'] = ';'.join(current_scene['relations'].keys())
                results[current_scene['key']] = current_scene
                current_scene = {}

        # events are normally in tag_ref1 and nouns in tag_ref2
        if prev_v is not None and not relation_processed:
            # ver caso o v seja a primeira relation na nova cena
            if current_scene == {}:  # first relation
                if v['rel_type'] in beginning_breakpoints:
                    if v['tag_ref1'] in events:  # verbo -> deve ser escrito em segundo
                        if v['tag_ref2'] in participants or (
                                v['tag_ref2'] in events and events[v['tag_ref2']]['Pos'] == 'Noun'):
                            current_scene['relations'] = {k: v}
                            current_scene['events'] = v['tag_ref1']
                            current_scene['participants'] = v['tag_ref2']
                    elif v['tag_ref1'] in participants or (
                            v['tag_ref1'] in events and events[v['tag_ref1']]['Pos'] == 'Noun'):
                        if v['tag_ref2'] in events:
                            current_scene['relations'] = {k: v}
                            current_scene['events'] = v['tag_ref2']
                            current_scene['participants'] = v['tag_ref1']
                else:
                    double_check.update({k: v})
            else:
                is_event = False
                is_participant = False
                scene_id = ''
                for kr, r in results.items():
                    if v['tag_ref1'] == list(r['relations'].values())[-1]['tag_ref1']:  # normally events
                        is_event = True
                        scene_id = kr
                    elif v['tag_ref1'] == list(r['relations'].values())[-1]['tag_ref2']:  # normally nouns
                        is_participant = True
                        scene_id = kr
                if v['tag_ref1'] == list(current_scene['relations'].values())[-1]['tag_ref1']:  # normally events
                    if v['tag_ref2'] in participants or (
                            v['tag_ref2'] in events and
                            events[v['tag_ref2']].get('Pos') == 'Noun'):
                        current_scene['relations'].update({k: v})
                        current_scene['participants'] += ';' + v['tag_ref2']
                elif v['tag_ref1'] == list(current_scene['relations'].values())[-1]['tag_ref2']:  # normally nouns
                    if v['tag_ref2'] in participants or (
                            v['tag_ref2'] in events and
                            events[v['tag_ref2']].get('Pos') == 'Noun'):
                        current_scene['relations'].update({k: v})
                        current_scene['participants'] += ';' + v['tag_ref2']
                    # falta locations mas isto se calhar é à parte porque há uma relaçao especifica para isto
                elif (is_event or is_participant) and scene_id != '':
                    if v['tag_ref2'] in participants or (
                            v['tag_ref2'] in events and
                            events[v['tag_ref2']].get('Pos') == 'Noun'):
                        results[scene_id]['relations'].update({k: v})
                        results[scene_id]['participants'] += ';' + v['tag_ref2']
                else:
                    double_check.update({k: v})

        if v['rel_type'] in end_breakpoints and current_scene != {}:
            if len(current_scene['relations']) == 1 and prev_v['tag_ref1'] in events and events[prev_v['tag_ref1']][
                'Class'] == 'Reporting':
                current_scene = {}
            else:
                current_scene['key'] = ';'.join(current_scene['relations'].keys())
                results[current_scene['key']] = current_scene
                current_scene = {}

        prev_k = k
        prev_v = v
        k = next_k
        v = next_v
        if list(semantic_roles_dict.keys())[-1] == k:
            finished = True
            # Last scene
            if current_scene != {}:
                if len(current_scene['relations']) == 1 and \
                        list(current_scene['relations'].values())[-1]['tag_ref1'] in events and \
                        events[list(current_scene['relations'].values())[-1]['tag_ref1']]['Class'] == 'Reporting':
                    current_scene = {}
                else:
                    current_scene['key'] = ';'.join(current_scene['relations'].keys())
                    results[current_scene['key']] = current_scene
                    current_scene = {}

    # For unordered relations, that are spread across the file
    for k, v in double_check.items():
        is_event = False
        is_participant = False
        scene_id = ''
        for kr, r in results.items():
            if v['tag_ref1'] == list(r['relations'].values())[-1]['tag_ref1']:  # normally events
                is_event = True
                scene_id = kr
            elif v['tag_ref1'] == list(r['relations'].values())[-1]['tag_ref2']:  # normally nouns
                is_participant = True
                scene_id = kr
            # falta locations mas isto se calhar é à parte porque há uma relaçao especifica para isto
        if (is_event or is_participant) and scene_id != '':
            if v['tag_ref2'] in participants or (
                    v['tag_ref2'] in events and
                    events[v['tag_ref2']].get('Pos') == 'Noun'):
                results[scene_id]['relations'].update({k: v})
                results[scene_id]['participants'] += ';' + v['tag_ref2']

    # Find the scenes to delete
    scenes_to_delete = []
    for k, r in results.items():
        if len(r['relations']) == 1 and \
                ((list(r['relations'].values())[-1]['tag_ref1'] in events and \
                  events[list(r['relations'].values())[-1]['tag_ref1']].get('Class') == 'Reporting') or
                 (list(r['relations'].values())[-1]['tag_ref2'] in events and \
                  events[list(r['relations'].values())[-1]['tag_ref2']].get('Class') == 'Reporting')):
            scenes_to_delete.append(r['key'])
        else:
            has_key_relation = False
            key_relations = ['SRLINK_agent', 'SRLINK_patient', 'SRLINK_pivot', 'SRLINK_result']
            agent_counter = 0  # if a scene has 2 or more agents then it's deleted
            for relation_key, relation_value in r['relations'].items():
                if relation_value['rel_type'] == 'SRLINK_agent':
                    has_key_relation = True
                    agent_counter += 1
                if relation_value['rel_type'] in key_relations:
                    has_key_relation = True
            if not has_key_relation or agent_counter > 1:
                scenes_to_delete.append(r['key'])
    for scene_to_del in scenes_to_delete:
        results.pop(scene_to_del)

    return results


def build_titles(extended_scenes):
    for scene_key, scene_value in extended_scenes.items():
        relations_list = list(scene_value['relations'].values())
        participants_mentioned = []
        for i, v in enumerate(relations_list):
            # First relation
            if i == 0:
                if v['tag_ref1'] in scene_value['events'].split(";"):  # verbo -> deve ser escrito em segundo
                    if v['tag_ref2'] in scene_value['participants'].split(";"):
                        scene_value['title'] = v['tag_ref2_value'].split(';')[0].strip() + ' ' + \
                                               v['tag_ref1_value'].split(';')[0].strip()
                        participants_mentioned.append(v['tag_ref2'])
                elif v['tag_ref1'] in scene_value['participants'].split(";"):
                    if v['tag_ref2'] in scene_value['events'].split(";"):
                        scene_value['title'] = v['tag_ref1_value'].split(';')[0].strip() + ' ' + \
                                               v['tag_ref2_value'].split(';')[0].strip()
                        participants_mentioned.append(v['tag_ref1'])
            # Next relation
            else:
                if v['tag_ref1'] == relations_list[i - 1]['tag_ref1']:  # normally events
                    if v['tag_ref2'] in scene_value['participants'] and v['tag_ref2'] not in participants_mentioned:
                        scene_value['title'] += ' ' + v['tag_ref2_value'].split(';')[0].strip()
                        participants_mentioned.append(v['tag_ref2'])
                elif v['tag_ref1'] == relations_list[i - 1]['tag_ref2']:  # normally nouns
                    if v['tag_ref2'] in scene_value['participants']:
                        scene_value['title'] += ' ' + v['tag_ref2_value'].split(';')[0].strip()
        if scene_value['title'][-3:] == 'que':
            scene_value['title'] = scene_value['title'][:len(scene_value['title']) - 4]

    # v['tag_ref2_value'] != 'que' -> tirar 'que's que estejam sozinhos
    return extended_scenes


def get_characters(all_entities):
    participants = filter_dict(all_entities.copy(), 'attribute', ['Participant'])
    characters = filter_dict(participants.copy(), 'Individuation_Domain', ['Individual', 'Set'])
    characters = filter_dict(characters.copy(), 'Participant_Type_Domain', ['Org', 'Per'])
    characters = filter_dict(characters.copy(), 'Lexical_Head', ['Noun'])
    characters_grouped = group_entities(characters.copy(), all_entities)

    return participants, characters, characters_grouped


def get_locations(participants, all_relations, all_entities):
    locations = filter_dict(participants.copy(), 'Individuation_Domain', ['Individual', 'Set'])
    location_type_domains = ['Fac', 'Path', 'Pl_water', 'Pl_celestial', 'Pl_mountain', 'Pl_civil', 'Pl_country',
                             'Pl_mount_range', 'Pl_capital', 'Pl_region', 'Pl_state', 'Loc']  # Loc não está documentado
    locations = filter_dict(locations.copy(), 'Participant_Type_Domain', location_type_domains)
    locations = filter_dict(locations.copy(), 'Lexical_Head', ['Noun'])

    qslink_relations = filter_dict(all_relations.copy(), 'rel_type', ['QSLINK_ground', 'QSLINK_figure'])
    spatial_relations = filter_dict(all_entities.copy(), 'attribute', ['Spatial_Relation'])

    # QSLINK_figure: entity or eventuality, QSLINK_ground: location
    spatial_information = group_qslinks_relations(qslink_relations, spatial_relations)
    #locations_grouped = group_entities(locations.copy(), all_entities)
    sr_loc_relations = filter_dict(all_relations.copy(), 'rel_type', ['SRLINK_location'])
    for kl, vl in sr_loc_relations.items():
        if vl['tag_ref1'] in locations:
            spatial_information.update({kl: {'QSLINK_figure': vl['tag_ref2'], 'QSLINK_ground': vl['tag_ref1']}})
        elif vl['tag_ref2'] in locations:
            spatial_information.update({kl: {'QSLINK_figure': vl['tag_ref1'], 'QSLINK_ground': vl['tag_ref2']}})

    return locations, spatial_information


def get_times(all_relations, events, all_times):
    tlinks_list = ['TLINK_before', 'TLINK_after', 'TLINK_includes', 'TLINK_isIncluded', 'TLINK_during',
                   'TLINK_simultaneous', 'TLINK_identity', 'TLINK_begins', 'TLINK_ends', 'TLINK_begunBy',
                   'TLINK_endedBy']
    tlinks_relations = filter_dict(all_relations.copy(), 'rel_type', tlinks_list)
    event_time_tlinks = get_event_time_tlinks(tlinks_relations, events, all_times)

    return event_time_tlinks


def get_sr_relations(f_parser, events, participants):
    sr_relations = get_semantic_relations(f_parser)
    # relations_to_consider = ['agent', 'partner', 'cause', 'patient', 'pivot', 'theme', 'patient', 'source', 'result',
    #                         'path', 'attribute']
    relations_to_consider = ['agent', 'partner', 'cause', 'patient', 'pivot', 'theme', 'beneficiary',
                             'source', 'goal', 'result', 'path', 'attribute', 'reason', 'purpose', 'amount']
    considered_sr_relations = filter_semantic_relations(relations_to_consider, sr_relations.copy())
    sr_with_info = link_sr_to_info(considered_sr_relations.copy(), participants, events)

    return sr_with_info


def get_extended_scenes(sr_with_info, participants, events):
    extended_scenes_assembled = assemble_extended_scenes(sr_with_info, participants, events)
    extended_scenes = build_titles(extended_scenes_assembled.copy())

    return extended_scenes


def brat2json(f, json_dir):
    print("-->", f)
    file_content = read_file(f)
    f_parser = file_parser(file_content)

    all_entities = get_all_entities(f_parser)
    all_relations = get_all_relations(f_parser)
    all_times = get_all_times(f_parser)

    (participants, characters, characters_grouped) = get_characters(all_entities)
    events = filter_dict(all_entities.copy(), 'attribute', ['Event'])
    (locations, spatial_information) = get_locations(participants, all_relations, all_entities)
    event_time_tlinks = get_times(all_relations, events, all_times)
    sr_with_info = get_sr_relations(f_parser, events, participants)
    extended_scenes = get_extended_scenes(sr_with_info, participants, events)

    txt_input_dir = '/'.join(f.split('/')[:len(f.split('/'))-1]) + '/'
    output_dir = '/'.join(f.split('/')[:len(f.split('/'))-2]) + '/' + json_dir
    # Create output dir if does not exist
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    write_jsons(characters_grouped, characters, f, txt_input_dir, output_dir, extended_scenes, events, locations,
                spatial_information, all_times, event_time_tlinks)


def bratdir2json(ann_dir, output_dir='output'):
    file_paths = get_files(ann_dir)  # get .ann files
    file_paths = [file_paths[x].replace('\\', '/') for x in range(0, len(file_paths))]
    for f in file_paths:
        brat2json(f, output_dir)


